'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from qsimov import QGate


def get_QFT(num_qubits):
    QFT = QGate(num_qubits, 0, f"QFT{num_qubits}")
    for i in range(num_qubits//2):
        QFT.add_operation("SWAP", targets=[i, num_qubits-i-1])
    for i in range(num_qubits):
        QFT.add_operation("H", targets=[i])
        for j in range(num_qubits - i - 1):
            QFT.add_operation(f"RUnity({j+2})", targets=[i], controls={j+i+1})
    return QFT


def get_swapped_QFT(num_qubits):
    QFT = QGate(num_qubits, 0, f"QFT{num_qubits}")
    for i in range(num_qubits):
        QFT.add_operation("H", targets=[i])
        for j in range(num_qubits - i - 1):
            QFT.add_operation(f"RUnity({j+2})", targets=[i], controls={j+i+1})
    return QFT



# -*- coding: utf-8 -*-
'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Base module for the Quantum Computing Simulator named QSimov.

Data Structures:
    Funmatrix: Functional Matrices related stuff in python
    QRegistry: Quantum Registry, base of all quantum related operations
    QSystem: Quantum System, preferred over QRegistry (can save a lot of space)
    QGate: Quantum Gate, built from elemental gates or other QGates
    QCircuit: Quantum Circuit, built from gates and measurements
    Measure: Data structure that represents a measurement in a circuit
    Condition: A condition to be evaluated after a measurement

Functions:
    setRandomSeed: sets the seed for measurements.
        Automatically called when this module is imported
    getGate: returns data about an elemental gate (not QGate)
    QEq: UNSTABLE. Compares two items
    getTruthTable: UNSTABLE. Returns the truth table of a gate

Data structures and functions marked as UNSTABLE might not work
Avoid using them as they might be removed in future versions
"""
import numpy as np
from qsimov.structures.qsystem import QSystem
from qsimov.structures.qgate import QGate

# np.zeros((h,w), dtype=complex) Inicializa una matriz de numeros complejos
# con alto h y ancho w
# La suma de matrices se realiza con +. A + B
# La multiplicacion por un escalar se hace con *. n * A
# Para multiplicar las matrices A y B se usa np.dot(A,B)
# El producto Kronecker de A y B esta definido con np.kron(A,B)


def getTruthTable(gate, ancilla=None, garbage=0, iterations=1):
    """Print the truth table of the given gate.

    You can set the ancilla bits to not include them in the table,
    with the list of values they must have.
    For example, if you have two 0s and one 1 as ancilla bits, ancilla[0,0,1].
    It always takes the last bits as the ancilla ones!
    The garbage=n removes the last n bits from the truth table,
    considering them garbage.
    For example, if you have 6 outputs and the last 4 outputs are garbage,
    only the value of the first two would be printed.
    Always removes the last n bits!
    """
    if (type(gate) == QGate):
        gate = gate.m
    num = int(np.log2(gate.getRows()))
    mesd = {}
    for iteration in range(iterations):
        for i in range(0, gate.getRows()):
            nbin = [int(x) for x in bin(i)[2:]]
            qinit = [0 for j in range(num - len(nbin))]
            qinit += nbin
            if ancilla is None or qinit[-len(ancilla):] == ancilla:
                qr = QSystem(len(qinit))
                qr.apply_gate(*[None if j == 0 else "X" for j in qinit])
                qr.apply_gate(gate)
                mes = qr.measure([1 for j in range(num-garbage)])
                if ancilla is not None:
                    ini = qinit[:-len(ancilla)]
                else:
                    ini = qinit
                if str(ini) not in mesd:
                    mesd[str(ini)] = np.zeros(num)
                mesd[str(ini)] = [x + y for x, y in zip(mesd[str(ini)], mes)]
    for k in mesd:
        for i in range(len(mesd[k])):
            mesd[k][i] /= iterations
            if (mesd[k][i] == 1.0 or mesd[k][i] == 0.0):
                mesd[k][i] = int(mesd[k][i])
        print(k + " -> " + str(["P(1)=" + str(v)
                                if type(v) != int and type(v) != int
                                else v
                                for v in mesd[k]]))
    return mesd


def QEq(q1, q2):
    """Return if q1 and q2 are equal and represented the same."""
    return np.array_equal(q1, q2) and str(q1) == str(q2)



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Module with functional matrices related stuff."""
import doki
import numpy as np


class Funmatrix(object):
    """Functional Matrices related stuff in python."""

    def __init__(self, fmatrix, name="Lazy Matrix", verbose=False):
        """Functional Matrix constructor.

        Positional arguments:
            fmatrix: pointer to C FunMatrix structure
        Keyworded arguments:
            name: string with the name of the matrix
        """
        self.m = fmatrix
        self.name = name
        self._shape = doki.funmatrix_shape(fmatrix, verbose)

    def __repr__(self):
        """Return the string representation of the FunMatrix."""
        return self.name

    def __str__(self):
        """Return the string representation of the FunMatrix."""
        return self.name

    def _get_single_item(self, i, j, verbose=False):
        """Return the item in (i, j) position of the matrix."""
        return doki.funmatrix_get(self.m, i, j, verbose)

    def __getitem__(self, key):
        """Return the items of the matrix specified in key."""
        rows, cols = self._shape

        if (type(key) == int):
            if (key >= rows):
                raise IndexError("index " + str(key) + " is out of bounds " +
                                 "for axis 0 with shape " + str(rows))
            if (key < 0):
                key = rows + key
            return np.array([self._get_single_item(key, j)
                             for j in range(cols)])
        elif (type(key) == slice):
            start = key.start if key.start is not None else 0
            if (start < 0):
                start = rows + start
            stop = key.stop if key.stop is not None else rows
            if (stop < 0):
                stop = rows + stop
            step = key.step if key.step is not None else 1
            return np.array([self.__getitem__(i)
                             for i in range(start, stop, step)])
        elif (type(key) == tuple):
            if (type(key[1]) is None):
                return self.__getitem__(key[0])
            elif (type(key[0]) is None):
                if (type(key[1]) == int):
                    if (key[1] >= cols):
                        raise IndexError("index " + str(key[1]) + " is out " +
                                         "of bounds for axis 1 with shape " +
                                         str(cols))
                    if (key[1] < 0):
                        key = (key[0], cols + key[1])
                    item_list = [np.around(self._get_single_item(i, key[1]),
                                           decimals=self.dec)
                                 for i in range(rows)]
                    return np.array(item_list)
                elif (type(key[1]) == slice):
                    start = key[1].start if key[1].start is not None else 0
                    if (start < 0):
                        start = cols + start
                    stop = key[1].stop if key[1].stop is not None else cols
                    if (stop < 0):
                        stop = cols + stop
                    step = key[1].step if key[1].step is not None else 1
                    return np.array([self.__getitem__(None, i)
                                     for i in range(start, stop, step)])
            elif (type(key[0]) == int and type(key[1]) == int):
                if (key[0] >= rows):
                    raise IndexError("index " + str(key[0]) + " is out of " +
                                     "bounds for axis 0 with shape " +
                                     str(rows))
                if (key[1] >= cols):
                    raise IndexError("index " + str(key[1]) + " is out of " +
                                     "bounds for axis 1 with shape " +
                                     str(cols))
                if (key[0] < 0):
                    key = (rows + key[0], key[1])
                if (key[1] < 0):
                    key = (key[0], cols + key[1])
                return self._get_single_item(key[0], key[1])
            elif (type(key[0]) == slice and type(key[1]) == int):
                if (key[1] >= cols):
                    raise IndexError("index " + str(key[1]) + " is out of " +
                                     "bounds for axis 1 with shape " +
                                     str(cols))
                start = key[0].start if key[0].start is not None else 0
                if (start < 0):
                    start = rows + start
                stop = key[0].stop if key[0].stop is not None else rows
                if (stop < 0):
                    stop = rows + stop
                step = key[0].step if key[0].step is not None else 1
                if (key[1] < 0):
                    key = (key[0], cols + key[1])
                return np.array([self._get_single_item(i, key[1])
                                 for i in range(start, stop, step)])
            elif (type(key[0]) == int and type(key[1]) == slice):
                if (key[0] >= rows):
                    raise IndexError("index " + str(key[0]) + " is out of " +
                                     "bounds for axis 0 with shape " +
                                     str(rows))
                start = key[1].start if key[1].start is not None else 0
                if (start < 0):
                    start = cols + start
                stop = key[1].stop if key[1].stop is not None else cols
                if (stop < 0):
                    stop = cols + stop
                step = key[1].step if key[1].step is not None else 1
                if (key[0] < 0):
                    key = (rows + key[0], key[1])
                return np.array([self._get_single_item(key[0], i)
                                 for i in range(start, stop, step)])
            elif (type(key[0]) == slice and type(key[1]) == slice):
                start0 = key[0].start if key[0].start is not None else 0
                if (start0 < 0):
                    start0 = rows + start0
                stop0 = key[0].stop if key[0].stop is not None else rows
                if (stop0 < 0):
                    stop0 = rows + stop0
                step0 = key[0].step if key[0].step is not None else 1
                start1 = key[1].start if key[1].start is not None else 0
                if (start1 < 0):
                    start1 = cols + start1
                stop1 = key[1].stop if key[1].stop is not None else cols
                if (stop1 < 0):
                    stop1 = cols + stop1
                step1 = key[1].step if key[1].step is not None else 1
                return np.array([[self._get_single_item(i, j)
                                  for i in range(start0, stop0, step0)]
                                 for j in range(start1, stop1, step1)])

    def __add__(self, other):
        """Return the result of adding self + other matrices."""
        if type(other) != Funmatrix:
            raise TypeError("unsupported operand type(s) for +: '" +
                            type(self).__name__ + "' and '" +
                            type(other).__name__ + "'")
        if self.shape != other.shape:
            raise ValueError("Both matrices must have the same shape")
        return Funmatrix(doki.funmatrix_add(self.m, other.m, False))

    def __sub__(self, other):
        """Return the result of substracting self - other matrices."""
        if type(other) != Funmatrix:
            raise TypeError("unsupported operand type(s) for -: '" +
                            type(self).__name__ + "' and '" +
                            type(other).__name__ + "'")
        if self.shape != other.shape:
            raise ValueError("Both matrices must have the same shape")
        return Funmatrix(doki.funmatrix_sub(self.m, other.m, False))

    def __mul__(self, other):
        """Calculate entity wise self * other or scalar product."""
        if type(other) == Funmatrix:
            if self.shape != other.shape:
                raise ValueError("Both matrices must have the same shape")
            return Funmatrix(doki.funmatrix_ewmul(self.m, other.m, False))
        elif np.isscalar(other):
            scalar = complex(other)
            return Funmatrix(doki.funmatrix_scalar_mul(self.m, scalar, False))
        else:
            raise TypeError("unsupported operand type(s) for *: '" +
                            type(self).__name__ + "' and '" +
                            type(other).__name__ + "'")

    def __rmul__(self, other):
        """Calculate entity wise other * self or scalar product."""
        if type(other) == Funmatrix:
            if self.shape != other.shape:
                raise ValueError("Both matrices must have the same shape")
            return Funmatrix(doki.funmatrix_ewmul(other.m, self.m, False))
        elif np.isscalar(other):
            scalar = complex(other)
            return Funmatrix(doki.funmatrix_scalar_mul(self.m, scalar, False))
        else:
            raise TypeError("unsupported operand type(s) for *: '" +
                            type(other).__name__ + "' and '" +
                            type(self).__name__ + "'")

    def __truediv__(self, other):
        """Calculate scalar division self / scalar."""
        if np.isscalar(other):
            scalar = complex(other)
            return Funmatrix(doki.funmatrix_scalar_div(self.m, scalar, False))
        else:
            raise TypeError("unsupported operand type(s) for /: '" +
                            type(self).__name__ + "' and '" +
                            type(other).__name__ + "'")

    def __matmul__(self, other):
        """Calculate matrix multiplication self . other."""
        if type(other) != Funmatrix:
            raise TypeError("unsupported operand type(s) for @: '" +
                            type(self).__name__ + "' and '" +
                            type(other).__name__ + "'")
        if self.shape[1] != other.shape[0]:
            raise ValueError("Matrix product is only defined when the number" +
                             " of columns of the first operand equals the" +
                             " number of rows of the second operand")
        return Funmatrix(doki.funmatrix_matmul(self.m, other.m, False))

    def __pow__(self, other):
        """Calculate matrix tensor product self x other or usual power."""
        if type(other) == Funmatrix:
            return Funmatrix(doki.funmatrix_kron(self.m, other.m, False))
        elif np.issubdtype(other, np.integer):
            exponent = int(other)
            if exponent < 0:
                raise ValueError("Negative exponents are not supported")
            res = self
            if exponent == 0:
                res = 1
            for i in range(exponent - 1):
                res = res @ self
            return res
        else:
            raise TypeError("unsupported operand type(s) for **: '" +
                            type(self).__name__ + "' and '" +
                            type(self).__name__ + "'")

    def transpose(self):
        """Return the transpose of this matrix."""
        return Funmatrix(doki.funmatrix_transpose(self.m, False))

    def dagger(self):
        """Return the conjugate (Hermitian) transpose of this matrix."""
        return Funmatrix(doki.funmatrix_dagger(self.m, False))

    def partial_trace(self, elem):
        """Return the partial trace of this matrix, after tracing out elem."""
        if self.shape[0] != self.shape[1]:
            raise ValueError("Partial trace undefined for non square matrices")
        num_elems = np.log2(self.shape[0])
        if not np.allclose(num_elems % 1, 0):
            raise NotImplementedError("Partial trace unsupported for " +
                                      "shapes that are not a power of 2")
        num_elems = int(num_elems)
        if not np.allclose(elem % 1, 0):
            raise ValueError("elem must be an integer")
        elem = int(elem)
        if elem < 0 or elem >= num_elems:
            raise ValueError(f"elem must be in [0, {num_elems - 1}] for" +
                             " this matrix")
        return Funmatrix(doki.funmatrix_partialtrace(self.m, elem, False))

    def trace(self):
        """Return the trace of this matrix."""
        if self.shape[0] != self.shape[1]:
            raise ValueError("Trace undefined for non square matrices")
        return doki.funmatrix_trace(self.m, False)

    def get_matrix(self):
        """Return self."""
        return self

    @property
    def shape(self):
        """Return tuple with (number of rows, number of columns)."""
        return self._shape



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Module with gate name parsing stuff.

This module has all name parsing stuff
"""

import re
import sympy as sp

from sympy.matrices import Matrix
from sympy.parsing.sympy_parser import parse_expr


__rep__ = re.compile(r"^([a-zA-Z0-9]+)(\(.*\))?(\-1)?$")


def parse_groups(groups):
    """Parse the result of get_groups function, passed as parameter."""
    errored = False
    g1 = groups[0]
    g4 = groups[2] is not None
    if groups[1] is not None:
        aux = groups[1][1:-1].split(",")
        g2 = len(aux)
        g3 = []
        for attr in aux:
            attr = parse_expr(attr)
            if not attr.is_number:
                print("Expression", attr, " is not a number")
                errored = True
                break
            g3.append(attr)
    else:
        g2 = 0
        g3 = None

    if not errored:
        return (g1, g2, g3, g4)
    else:
        return None


def get_groups(str_gate):
    """Get matching groups using __rep__ regular expression."""
    res = __rep__.match(str_gate)
    return parse_groups(res.groups()) if res is not None else None


def Hadamard():
    """Return sympy matrix with Hadamard gate for 1 qubit."""
    return (sp.sqrt(2) / 2) * Matrix([[1, 1], [1, -1]])


def PauliX():
    """Return sympy matrix with Pauli X gate (aka NOT)."""
    return Matrix([[0, 1], [1, 0]])


def PauliY():
    """Return sympy matrix with PauliY gate."""
    return Matrix([[0, -1j], [1j, 0]])


def PauliZ():
    """Return sympy matrix with PauliZ gate."""
    return Matrix([[1, 0], [0, -1]])


def SqrtX():
    """Return sympy matrix with sqrt(NOT) gate."""
    return Matrix([[0.5 + 0.5j, 0.5 - 0.5j], [0.5 - 0.5j, 0.5 + 0.5j]])


def P(angle):
    """Return sympy matrix with PhaseChange gate."""
    _R = sp.eye(2)
    _R[1, 1] = sp.exp(sp.I * angle)
    return _R


def R(theta, phi):
    """Return sympy matrix with R gate."""
    _R = sp.eye(2) * sp.cos(theta / 2)
    _isinth2 = -sp.I * sp.sin(theta / 2)
    _R[0, 1] = sp.exp(-sp.I * phi) * _isinth2
    _R[1, 0] = sp.exp(sp.I * phi) * _isinth2
    return _R


def RX(angle):
    """Return sympy matrix with rotation gate around X axis."""
    gate = Matrix([[0, 0], [0, 0]])
    cosan = sp.cos(angle / 2)
    sinan = -1j * sp.sin(angle / 2)
    gate[0, 0] = cosan
    gate[0, 1] = sinan
    gate[1, 0] = sinan
    gate[1, 1] = cosan
    return gate


def RY(angle):
    """Return sympy matrix with rotation gate around Y axis."""
    gate = Matrix([[0, 0], [0, 0]])
    cosan = sp.cos(angle / 2)
    sinan = sp.sin(angle / 2)
    gate[0, 0] = cosan
    gate[1, 1] = cosan
    gate[0, 1] = -sinan
    gate[1, 0] = sinan
    return gate


def RZ(angle):
    """Return sympy matrix with rotation gate around Z axis."""
    gate = Matrix([[0, 0], [0, 0]])
    gate[0, 0] = sp.cos(-angle / 2) + sp.sin(-angle / 2) * 1j
    gate[1, 1] = sp.cos(angle / 2) + sp.sin(angle / 2) * 1j
    return gate


def RUnity(n):
    """Return sympy matrix with nth root of unity rotation gate."""
    return P(2*sp.pi/(2**n))


def HalfDeutsch(angle):
    """Return sympy matrix with a portion of the Deutsch gate.

    This gate, when double controlled, is called Deutsch gate.
    """
    gate = Matrix([[0, 0], [0, 0]])
    cosan = sp.cos(angle) * sp.I
    sinan = sp.sin(angle)
    gate[0, 0] = cosan
    gate[0, 1] = sinan
    gate[1, 0] = sinan
    gate[1, 1] = cosan
    return gate


def U(th, ph, la):
    """Return sympy matrix with U(θ, φ, λ) gate (IBM)."""
    gate = Matrix([[0, 0], [0, 0]])
    costh2 = sp.cos(th / 2)
    sinth2 = sp.sin(th / 2)
    gate[0, 0] = costh2
    gate[0, 1] = -sp.exp(sp.I * la) * sinth2
    gate[1, 0] = sp.exp(sp.I * ph) * sinth2
    gate[1, 1] = sp.exp(sp.I * (ph + la)) * costh2
    return gate


def U3(th, ph, la):
    """Return sympy matrix with U3(θ, φ, λ) gate (IBM)."""
    print("[WARNING] This gate has been deprecated in OpenQASM standard. Use U(θ, φ, λ) instead.")
    return U(th, ph, la)


def U2(ph, la):
    """Return sympy matrix with U2 gate (IBM)."""
    print("[WARNING] This gate has been deprecated in OpenQASM standard. Use U(π/2, φ, λ) instead.")
    return U(sp.pi / 2, ph, la)


def U1(angle):
    """Return sympy matrix with U1 gate (IBM)."""
    print("[WARNING] This gate has been deprecated in OpenQASM standard. Use U(0, 0, λ) or P(λ) instead.")
    return P(angle)


def SWAP():
    """Return sympy matrix with SWAP gate."""
    gate = sp.zeros(4, 4)
    gate[0, 0] = 1
    gate[1, 2] = 1
    gate[2, 1] = 1
    gate[3, 3] = 1
    return gate


def iSWAP():
    """Return sympy matrix with iSWAP gate."""
    gate = sp.zeros(4, 4)
    gate[0, 0] = 1
    gate[1, 2] = 1j
    gate[2, 1] = 1j
    gate[3, 3] = 1
    return gate


def fSWAP():
    """Return sympy matrix with fermionic swap fSWAP gate."""
    gate = sp.zeros(4, 4)
    gate[0, 0] = 1
    gate[1, 2] = 1
    gate[2, 1] = 1
    gate[3, 3] = -1
    return gate


def sqrtSWAP():
    """Return sympy matrix with sqrt(SWAP) gate."""
    gate = sp.zeros(4, 4)
    gate[0, 0] = 1
    gate[1, 1] = 0.5 + 0.5j
    gate[1, 2] = 0.5 - 0.5j
    gate[2, 1] = 0.5 - 0.5j
    gate[2, 2] = 0.5 + 0.5j
    gate[3, 3] = 1
    return gate


def xx(angle):
    """Return sympy matrix with Ising Coupling XX gate. AKA Mølmer–Sørensen gate"""
    gate = sp.eye(4)
    phi2 = angle / 2
    gate = gate * sp.cos(phi2)
    _isinphi2 = -1j * sp.sin(phi2)
    gate[0, 3] = _isinphi2
    gate[1, 2] = _isinphi2
    gate[2, 1] = _isinphi2
    gate[3, 0] = _isinphi2
    return gate


def yy(angle):
    """Return sympy matrix with Ising Coupling YY gate."""
    gate = sp.eye(4)
    phi2 = angle / 2
    gate = gate * sp.cos(phi2)
    isinphi2 = 1j * sp.sin(phi2)
    gate[0, 3] = isinphi2
    gate[1, 2] = -isinphi2
    gate[2, 1] = -isinphi2
    gate[3, 0] = isinphi2
    return gate


def zz(angle):
    """Return sympy matrix with Ising Coupling ZZ gate."""
    gate = sp.eye(4)
    phi2 = angle / 2
    cosphi2 = sp.cos(phi2)
    isinphi2 = 1j * sp.sin(phi2)
    gate[0, 0] = cosphi2 - isinphi2
    gate[1, 1] = cosphi2 + isinphi2
    gate[2, 2] = cosphi2 + isinphi2
    gate[3, 3] = cosphi2 - isinphi2
    return gate


def xy(angle):
    """Return sympy matrix with Ising Coupling ZZ gate."""
    gate = sp.eye(4)
    phi2 = angle / 2
    cosphi2 = sp.cos(phi2)
    _isinphi2 = -1j * sp.sin(phi2)
    gate[1, 2] = _isinphi2
    gate[1, 1] = cosphi2
    gate[2, 2] = cosphi2
    gate[2, 1] = _isinphi2
    return gate


_gate_alias = {}
_gate_alias["x"] = "X"
_gate_alias["not"] = "X"
_gate_alias["sqrtnot"] = "SqrtX"
_gate_alias["sqrtx"] = "SqrtX"
_gate_alias["v"] = "SqrtX"
_gate_alias["y"] = "Y"
_gate_alias["z"] = "Z"
_gate_alias["rx"] = "RX"
_gate_alias["ry"] = "RY"
_gate_alias["rz"] = "RZ"
_gate_alias["r"] = "R"
_gate_alias["p"] = "P"
_gate_alias["phaseshift"] = "P"
_gate_alias["phasechange"] = "P"
_gate_alias["runity"] = "RootPhase"
_gate_alias["rootphase"] = "RootPhase"
_gate_alias["h"] = "H"
_gate_alias["u"] = "U"
_gate_alias["u3"] = "U3"
_gate_alias["u2"] = "U2"
_gate_alias["u1"] = "U1"
_gate_alias["d"] = "HalfDeutsch"
_gate_alias["deutsch"] = "HalfDeutsch"
_gate_alias["halfdeutsch"] = "HalfDeutsch"
_gate_alias["partialdeutsch"] = "HalfDeutsch"
_gate_alias["xx"] = "XX"
_gate_alias["rxx"] = "XX"
_gate_alias["yy"] = "YY"
_gate_alias["ryy"] = "YY"
_gate_alias["zz"] = "ZZ"
_gate_alias["rxx"] = "ZZ"
_gate_alias["xy"] = "XY"
_gate_alias["rxy"] = "XY"
_gate_alias["swap"] = "SWAP"
_gate_alias["iswap"] = "iSWAP"
_gate_alias["fswap"] = "fSWAP"
_gate_alias["sqrtswap"] = "SqrtSWAP"

# min_args, max_args, has_invert_arg, is_self_invert
_gate_data = {}
_gate_func = {}
_gate_data["X"] = (0, 0, True)
_gate_func["X"] = PauliX
_gate_data["Y"] = (0, 0, True)
_gate_func["Y"] = PauliY
_gate_data["Z"] = (0, 0, True)
_gate_func["Z"] = PauliZ
_gate_data["H"] = (0, 0, True)
_gate_func["H"] = Hadamard
_gate_data["SqrtX"] = (0, 0, False)
_gate_func["SqrtX"] = SqrtX
_gate_data["RX"] = (1, 1, False)
_gate_func["RX"] = RX
_gate_data["RY"] = (1, 1, False)
_gate_func["RY"] = RY
_gate_data["RZ"] = (1, 1, False)
_gate_func["RZ"] = RZ
_gate_data["R"] = (2, 2, False)
_gate_func["R"] = R
_gate_data["P"] = (1, 1, False)
_gate_func["P"] = P
_gate_data["RootPhase"] = (1, 1, False)
_gate_func["RootPhase"] = RUnity
_gate_data["U"] = (3, 3, False)
_gate_func["U"] = U
_gate_data["U3"] = (3, 3, False)
_gate_func["U3"] = U3
_gate_data["U2"] = (2, 2, False)
_gate_func["U2"] = U2
_gate_data["U1"] = (1, 1, False)
_gate_func["U1"] = U1
_gate_data["HalfDeutsch"] = (1, 1, False)
_gate_func["HalfDeutsch"] = HalfDeutsch
_gate_data["XX"] = (1, 1, False)
_gate_func["XX"] = xx
_gate_data["YY"] = (1, 1, False)
_gate_func["YY"] = yy
_gate_data["ZZ"] = (1, 1, False)
_gate_func["ZZ"] = zz
_gate_data["XY"] = (1, 1, False)
_gate_func["XY"] = xy
_gate_data["SWAP"] = (0, 0, True)
_gate_func["SWAP"] = SWAP
_gate_data["iSWAP"] = (0, 0, False)
_gate_func["iSWAP"] = iSWAP
_gate_data["fSWAP"] = (0, 0, False)
_gate_func["fSWAP"] = fSWAP
_gate_data["SqrtSWAP"] = (0, 0, False)
_gate_func["SqrtSWAP"] = sqrtSWAP


def get_available_gates():
    return set(_gate_data.keys())


def get_gate_aliases():
    return _gate_alias.copy()


def get_gate_data(gateraw):
    """Get the data of the gate associated with the given string."""
    gate = None
    if type(gateraw) == str:
        groups = get_groups(gateraw)
        if groups is not None:
            alias, nargs, args, invert = groups
            alias = alias.lower()
            if alias in _gate_alias:
                gatename = _gate_alias[alias]
                if gatename in _gate_data:
                    minargs, maxargs, self_inv = _gate_data[gatename]
                    if self_inv:  # If gate . gate = identity
                        invert = False
                    if minargs <= nargs <= maxargs:  # Adoro Python
                        gate = (gatename, args, invert, self_inv)
                    else:
                        if gatename == "R" and nargs == 1:
                            print("[ERROR] The old R gate is now called P in OpenQASM standard. Thus, we did the same.")
                        raise ValueError(f"{gatename} gate number of args" +
                                         f" must be between {minargs} and {maxargs}." +
                                         f" Got {nargs} from {gateraw}")
                else:
                    raise ValueError("Couldn't find data for gate " + gatename)
            else:
                raise ValueError(gateraw + " is not a gate." +
                                 " Have you already added it?")
        else:
            raise ValueError(gateraw + " can't be used with QSimovAPI")
    else:
        raise ValueError("You can only use a string!")
    return gate



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import numpy as np
import qsimov.connectors.parser as gt
import sympy as sp

from sympy.functions import conjugate
from sympy.matrices import Matrix
from sympy.physics.quantum import TensorProduct


_pauli = [None for i in range(3)]
_pauli[0] = gt.PauliX() # σx
_pauli[1] = gt.PauliY() # σy
_pauli[2] = gt.PauliZ() # σz

# Unitary vectors
_u = [None for i in range(3)]
_u[0] = Matrix([[1, 0, 0]])
_u[1] = Matrix([[0, 1, 0]])
_u[2] = Matrix([[0, 0, 1]])


def gate_new(num_qubits, matrix, verbose):
    if num_qubits > 1:
        raise ValueError("Gates for bloch spheres can affect at most one single qubit")
    return BRotation(matrix, verbose)


def gate_get(gate, verbose):
    return gate.O


def registry_new(num_qubits, verbose):
    return BRegistry(num_qubits, verbose=verbose)


def registry_clone(registry, num_threads, verbose):
    newr = BRegistry(registry.nq, verbose=verbose)
    for id in range(registry.nq):
        newr.r[id].v = Matrix([registry.r[id].v[0, 0],
                               registry.r[id].v[1, 0],
                               registry.r[id].v[2, 0]])
    return newr


def registry_del(registry, verbose):
    del registry


def registry_apply(registry, gate, target_list, control_set, anticontrol_set, num_threads, verbose):
    if control_set != None and len(control_set) > 0:
        raise ValueError("Gates for bloch spheres can have controls")
    if anticontrol_set != None and len(anticontrol_set) > 0:
        raise ValueError("Gates for bloch spheres can have anticontrols")
    return registry.apply_gate(gate, targets=target_list, verbose=verbose)


def registry_join(most_registry, least_registry, num_threads, verbose):
    newr = BRegistry(most_registry.nq + least_registry.nq, init=False, verbose=verbose)
    newr.r = most_registry.r + least_registry.r


def registry_measure(registry, mask, roll_list, num_threads, verbose):
    targets = [id for id in range(registry.nq) if bool(mask & (1 << id))]
    registry.measure(targets=targets, rands=roll_list, verbose=verbose)


def registry_prob(registry, qubit_id, num_threads, verbose):
    return registry.r[qubit_id].v[2, 0]


class BRotation(object):
    def __init__(self, matrix, verbose=False):
        """https://quantumcomputing.stackexchange.com/questions/16533/can-i-find-the-axis-of-rotation-for-any-single-qubit-gate"""
        # e^iα = sqrt(det(O))
        self.verbose = verbose
        self.O = matrix
        if verbose:
            print("Creating gate for matrix", matrix)
        eia = sp.sqrt(sp.det(self.O))
        alpha = sp.atan2(sp.im(eia), sp.re(eia))
        e_ia = conjugate(eia)
        self.theta = 2 * sp.acos(e_ia * sp.trace(self.O) / 2)
        _2isint2 = -2j * sp.sin(self.theta / 2)
        self.n = Matrix([[(e_ia * sp.trace(self.O @ pw)) / _2isint2 for pw in _pauli]])

    def __str__(self):
        return f"Rotation θ = {self.theta}, axis = {self.n}"

    def __repr__(self):
        return str(self)

    def rotation_matrix(self):
        cost = sp.cos(self.theta)
        sint = sp.sin(self.theta)
        u = self.n
        auxu = [TensorProduct(u.cross(_u[i]), _u[i]) for i in range(3)]
        ux = (auxu[0] + auxu[1] + auxu[2]).reshape(3, 3)
        r = cost * sp.eye(3) + sint * ux + (1 - cost) * TensorProduct(u, u).reshape(3,3)
        return r


class QuBit(object):
    def __init__(self):
        self.v = Matrix([0, 0, 1])

    def __str__(self):
        return str(self.v.transpose())

    def __repr__(self):
        return str(self)

    def apply_gate(self, rot, verbose=False):
        m = rot.rotation_matrix()
        qb = QuBit()
        qb.v = m @ self.v
        if verbose:
            print("Prev:", self.v)
            print("Post:", qb.v)
        return qb

    def measure(self, rand=None):
        zproj = self.v[2, 0]
        qb = QuBit()
        res = 0
        odds = 1 - ((zproj + 1) / 2)
        if rand is None:
            rand = np.random.rand()
        if rand < odds:
            qb.v = Matrix([0, 0, -1])
            res = 1
        return qb, res


class BRegistry(object):
    def __init__(self, nq, init=True, verbose=False):
        self.nq = nq
        self.r = None
        if init:
            self.r = tuple(QuBit() for i in range(nq))

    def apply_gate(self, gate, targets=0, verbose=False):
        if type(targets) == int:
            targets = [targets]
        t_set = set(targets)
        if len(t_set) != len(targets):
            raise ValueError("Targets has repeated elements")
        rot = BRotation(gate, verbose=verbose)
        if verbose:
            print("Rotation matrix:", rot.rotation_matrix())
        newr = BRegistry(self.nq, False)
        newr.r = tuple(self.r[i].apply_gate(rot, verbose=verbose) if i in t_set else self.r[i] for i in range(self.nq))
        return newr

    def measure(self, targets=0, rands=None, verbose=False):
        if type(targets) == int:
            targets = [targets]
        t_set = set(targets)
        if len(t_set) != len(targets):
            raise ValueError("Targets has repeated elements")
        if rands is None:
            rands = np.random.rand(len(t_set))
        if len(rands) != len(t_set):
            raise ValueError("Rands has to have the same number of elements as targets")
        newr = BRegistry(self.nq, False)
        rands = [rands.pop(0) if i in t_set else None for i in range(self.nq)]
        aux = tuple(self.r[i].measure(rands[i]) if i in t_set else self.r[i] for i in range(self.nq))
        newr.r = tuple(aux[i][0] if i in t_set else aux[i] for i in range(self.nq))
        res = tuple(aux[i][1] if i in t_set else None for i in range(self.nq))
        return newr, res



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Deutsch-Jozsa with n qubit."""
from qsimov import QGate

import sympy as sp


def first_column_gate(num_qubits, vector):
    """Create a gate with normalized vector as first matrix column."""
    gates = get_gates(vector, num_qubits)
    # print("Gates:", gates)
    # print("V:", vector)
    vector_oracle = QGate(num_qubits, "Vector oracle")
    if len(gates) > 0:
        for angle, qid, controls, anticontrols in gates:
            vector_oracle.add_operation("U(" + str(angle * 2) + ",0,pi)",
                                        targets=qid, controls=controls,
                                        anticontrols=anticontrols)
    # print("Vector:", vector)
    # print("Gates:", gates)
    return vector_oracle


def get_gates(vector, num_qubits):
    """Get all gates from the vector."""
    gates = []
    for j in range(1, num_qubits + 1):
        for i in range(2**(j - 1)):
            control_string = ""
            format_string = "{:0" + str(j - 1) + "b}"
            if j > 1:
                control_string = format_string.format(i)
            angle = get_angle_from_j(vector, j, num_qubits, control_string)
            controls = {num_qubits - i - 1
                        for i in range(len(control_string))
                        if control_string[i] == "1"}
            anticontrols = {num_qubits - i - 1
                            for i in range(len(control_string))
                            if control_string[i] == "0"}
            if angle != 0:
                gates.append((angle, num_qubits - j, controls, anticontrols))

    return gates


def get_angle_from_j(vector, j, n, control_string):
    """Get the rotation angle."""
    numerator = 0
    denominator = 0
    is_last = j == n
    last_imag = False
    if not is_last:
        for id in get_ids_from_j(j, n, control_string, True):
            numerator += abs(vector[id])**2
        for id in get_ids_from_j(j, n, control_string, False):
            denominator += abs(vector[id])**2
    else:
        numerator = vector[get_ids_from_j(j, n, control_string, True)[0]]
        denominator = vector[get_ids_from_j(j, n, control_string, False)[0]]
        if numerator.imag != 0 or denominator.imag != 0:
            last_imag = True
            numerator = abs(numerator)**2
            denominator = abs(denominator)**2
    angle = 0
    if numerator != 0 and denominator != 0:
        if not is_last or last_imag:
            numerator = sp.sqrt(numerator)
            denominator = sp.sqrt(denominator)
        angle = sp.atan2(numerator, denominator).evalf()
    else:  # 0/0
        subvector = vector
        for i in range(len(control_string)):
            bit_value = int(control_string[i])
            bit_id = n - i - 1
            subvector = get_bit_elements(subvector, bit_id, bit_value)
        subvector = get_bit_elements(subvector, n - j, 0)
        if all([subvector[i] == 0 for i in range(len(subvector))]):
            angle = (sp.pi/2).evalf()
        else:
            angle = 0

    return angle


def get_ids_from_j(j, n, control_string, is_numerator):
    """Get the id of the vector items."""
    static_str = control_string + str(int(is_numerator))
    bits = n - j
    if bits > 0:
        bit_format = "{:0" + str(bits) + "b}"
        ids = [int(static_str + bit_format.format(i), 2)
               for i in range(2**bits)]
    else:
        ids = [int(static_str, 2)]
    return ids


def get_bit_elements(vector, bit, value):
    """Return a list with the elements of vector that has bit with value."""
    subvector = []
    found = 0
    max_found = len(vector)//2
    for i in range(len(vector)):
        bin_i_rev = ("{:0" + str(2**bit) + "b}").format(i)[::-1]
        if int(bin_i_rev[bit]) == value:
            subvector.append(vector[i])
            found += 1
            if found == max_found:
                break
    return subvector



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import numpy as np
from abc import abstractmethod
from collections.abc import Iterable
from qsimov.structures.qbase import QBase
from qsimov.structures.simple_gate import SimpleGate


class QStructure(QBase):
    @abstractmethod
    def __init__(self, num_qubits, data=None, doki=None, verbose=False):
        pass

    @abstractmethod
    def apply_gate(self, gate, targets=None, controls=None, anticontrols=None,
                   num_threads=-1):
        pass

    @abstractmethod
    def measure(self, ids, random_generator=np.random.rand):
        pass

    @abstractmethod
    def prob(self, id):
        pass

    @abstractmethod
    def get_state_size(self):
        pass

    @abstractmethod
    def get_state(self, key=None, canonical=False):
        pass

    @abstractmethod
    def get_data(self):
        pass

    @abstractmethod
    def get_classic(self, id):
        pass

    @abstractmethod
    def clone(self, num_threads=-1):
        pass

    @abstractmethod
    def free(self):
        pass

    @abstractmethod
    def get_bloch_coords(self, key=None):
        pass

    def bloch(self, key=None):
        """Return matplotlib bloch sphere."""
        all_coords = self.get_bloch_coords(key=key)
        if type(all_coords) != list:
            all_coords = [all_coords]
        from qsimov.utils.bloch import draw_bloch_sphere
        figs = [draw_bloch_sphere(float(coords[0]), float(coords[1])) if coords is not None
                else None
                for coords in all_coords]
        if key is not None and type(key) != slice:
            figs = figs[0]
        return figs


def _get_op_data(num_qubits, num_bits, gate, targets, c_targets, outputs,
                 controls, anticontrols, c_controls, c_anticontrols,
                 empty=False):
    """Do basic error checking for arguments and return them."""
    targets = _get_qubit_set(num_qubits, targets, True, "targets")
    c_targets = _get_qubit_set(num_bits, c_targets, True, "classic targets")
    if gate is not None:
        num_c_targets = 0
        if type(gate) == str:
            gate = SimpleGate(gate)
        elif type(gate) != SimpleGate:
            num_c_targets = gate.num_bits
        num_targets = gate.num_qubits
        if len(targets) == 0:  # By default we use the least significant qubits
            targets = [i for i in range(num_targets)]
        if len(c_targets) == 0:  # By default we use the least significant bits
            c_targets = [i for i in range(num_c_targets)]
        if len(targets) != num_targets:
            raise ValueError(f"Specified gate is for {num_targets} qubits." +
                             f" {len(targets)} qubit ids given")
        if len(c_targets) != num_c_targets:
            raise ValueError(f"Specified gate is for {num_c_targets} bits." +
                             f" {len(c_targets)} bit ids given")
    controls = _get_qubit_set(num_qubits, controls, False, "controls")
    c_controls = _get_qubit_set(num_bits, c_controls,
                                False, "classic controls")
    anticontrols = _get_qubit_set(num_qubits, anticontrols,
                                  False, "anticontrols")
    c_anticontrols = _get_qubit_set(num_bits, c_anticontrols,
                                    False, "classic anticontrols")
    outputs = _get_qubit_set(num_bits, outputs, True, "outputs")
    _check_no_intersection(targets, controls, anticontrols)
    _check_no_intersection(c_targets, c_controls, c_anticontrols, True)
    if gate is None:
        if not empty:
            if len(outputs) != len(targets):
                raise ValueError(f"Expected {len(targets)} output bits." +
                                 f" {len(outputs)} ids given")
            if len(controls) + len(anticontrols) > 0:
                raise ValueError("Measures can only be controlled by bits")
    elif len(outputs) != 0:
        raise ValueError("Gate applications can't have classical outputs")

    return {"gate": gate,
            "targets": targets, "c_targets": c_targets, "outputs": outputs,
            "controls": controls, "anticontrols": anticontrols,
            "c_controls": c_controls, "c_anticontrols": c_anticontrols}


def _check_no_intersection(targets, controls, anticontrols, classic=False):
    """Raise an exception if any qubit id is used more than once."""
    aux = ""
    if classic:
        aux = "classic "
    if len(controls.intersection(targets)) > 0:
        raise ValueError(f"A {aux}target cannot also be a control")
    if len(anticontrols.intersection(targets)) > 0:
        raise ValueError(f"A {aux}target cannot also be an anticontrol")
    if len(controls.intersection(anticontrols)) > 0:
        raise ValueError(f"A {aux}control cannot also be an anticontrol")


def _get_qubit_set(max_qubits, raw_ids, sorted, name):
    """Get a set or sorted set (list) of qubit ids from raw_ids."""
    if raw_ids is None:
        if sorted:
            return []
        else:
            return set()
    if not isinstance(raw_ids, Iterable):
        raw_ids = [raw_ids]
    num_ids = len(raw_ids)
    ids_check = all([np.allclose(qubit_id % 1, 0)
                     and qubit_id < max_qubits
                     and qubit_id >= 0
                     for qubit_id in raw_ids])
    if not ids_check:
        raise ValueError(f"Invalid id found in {name}")
    if sorted:
        id_list = [int(raw_ids[i]) for i in range(num_ids)]
    else:
        id_list = [raw_id for raw_id in raw_ids]
    id_set = set(id_list)
    if num_ids != len(id_set):  # Check duplicates
        raise ValueError(f"{name} list cannot have duplicated ids")
    if sorted:
        return id_list
    return id_set


def _get_key_with_defaults(key, size, def_start, def_stop, def_step):
    if key is None:
        key = slice(def_start, def_stop, def_step)
    if type(key) != slice and np.allclose(key % 1, 0):
        key = int(key)
        if (key < 0):
            key = size + key
        if (key >= size or key < 0):
            raise IndexError(f"index {key} is out of bounds " +
                             f"for axis 0 with shape {size}")
        key = slice(key, key + 1, 1)
    if type(key) != slice:
        raise ValueError("key must be an index or a slice")

    start = key.start if key.start is not None else def_start
    if (start < 0):
        start = size + start
        if (start < 0):
            start = 0
    stop = key.stop if key.stop is not None else def_stop
    if (stop < 0):
        stop = size + stop
        if (stop < 0):
            stop = 0
    step = key.step if key.step is not None else def_step

    return (start, stop, step)



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Module that provides the lowest level data structure: QRegistry.

Data Structures:
    QRegistry: Quantum Registry, base of all quantum related operations

Functions:
    superposition: join two registries into one by calculating tensor product.
"""
import numpy as np
import sympy as sp

from sympy.functions.elementary.complexes import arg
from qsimov.structures.funmatrix import Funmatrix
from qsimov.structures.simple_gate import SimpleGate
from qsimov.structures.qstructure import QStructure, _get_op_data, \
                                         _get_qubit_set, _get_key_with_defaults
from qsimov.connectors.qsimovapi import apply_design


class QRegistry(QStructure):
    """Quantum Registry, base of all quantum related operations."""

    def __init__(self, num_qubits, data=None, doki=None, verbose=False):
        """Initialize QRegistry to state 0.

        num_qubits -> number of QuBits in the registry.
        """
        if doki is None:
            import doki
        self.doki = doki
        if data is None:
            if num_qubits is not None:
                if num_qubits > 0:
                    self.reg = self.doki.registry_new(num_qubits, verbose)
                    self.qubit_map = {i: i for i in range(num_qubits)}
                    self.classic_vals = {}
                    self.num_qubits = num_qubits
                    self.num_bits = 0
                    self.size = 2**num_qubits
                else:
                    raise ValueError("num_qubits must be greater than 0")
            else:
                self.reg = None
                self.num_qubits = 0
                self.num_bits = 0
                self.size = 0
                self.qubit_map = {}
                self.classic_vals = {}
        else:
            self.reg = self.doki.registry_new_data(data["num_qubits"], data["reg"], verbose)
            self.num_qubits = data["num_qubits"]
            self.num_bits = data["num_bits"]
            self.size = data["size"]
            self.qubit_map = data["qubit_map"]
            self.classic_vals = data["classic_vals"]
        self.verbose = verbose

    def __del__(self):
        """Clean after deletion."""
        self.free()

    def get_data(self):
        return {"reg": self.get_state(),
                "num_qubits": self.num_qubits,
                "num_bits": self.num_bits,
                "size": self.size,
                "qubit_map": self.qubit_map,
                "classic_vals": self.classic_vals}

    def get_classic(self, id):
        """Return classic bit value (if qubit has been measured)."""
        if id in self.classic_vals:
            return self.classic_vals[id]
        return None

    def free(self):
        """Release memory held by the QRegistry."""
        if self.reg is not None:
            self.doki.registry_del(self.reg, self.verbose)
            self.reg = None
            self.num_qubits = 0
            self.size = 0

    def get_state_size(self):
        """Return the number of elements in the state vector."""
        return self.size

    def get_num_qubits(self):
        """Return the number of qubits in this registry."""
        return self.num_qubits + self.num_bits

    def measure(self, ids, random_generator=np.random.rand, num_threads=-1):
        """Measure specified qubits of this registry and collapse.

        Positional arguments:
            ids -> List of QuBit ids that have to be measured
        Keyworded arguments:
            random_generator -> function without arguments that returns
                                a random real number in [0, 1)
        Return:
            List with the value obtained after each measure
        """
        ids_set = _get_qubit_set(self.get_num_qubits(),
                                 ids, False, "ids")
        if len(ids_set) == 0:
            raise ValueError('ids cannot be empty')
        num_measured = len(ids_set)
        mask = 0
        rolls = [random_generator() for i in range(num_measured)]
        for id in ids_set:
            if id not in self.qubit_map:
                raise ValueError(f"Id {id} has already been measured")
            mask += 2**self.qubit_map[id]
        raw_reg, mess = self.doki.registry_measure(self.reg, mask, rolls,
                                                   num_threads, self.verbose)
        mess = mess[::-1]
        final_mess = [mess[self.qubit_map[id]] if id in ids_set
                      else None
                      for id in range(self.num_qubits + self.num_bits)]
        rem_qubits = self.num_qubits - num_measured
        if raw_reg is not None:
            # If rem_qubits is 0 then raw_reg must be None
            if rem_qubits == 0:
                raise RuntimeError("Zero qubit registry found while " +
                                   "measuring! Please report this bug.")
        new_reg = QRegistry(None, doki=self.doki)
        new_reg.num_qubits = rem_qubits
        new_reg.num_bits = self.num_bits + num_measured
        new_reg.size = 2**rem_qubits
        new_reg.reg = raw_reg
        new_reg.verbose = self.verbose
        # Copying maps
        old_keys = [k for k in self.qubit_map if k not in ids_set]
        # print("DEBUG -> old_keys:", old_keys)
        new_reg.qubit_map = {old_keys[i]: i
                             for i in range(len(old_keys))}
        num_total = new_reg.num_qubits + new_reg.num_bits
        new_reg.classic_vals = {id: self.classic_vals[id]
                                if id in self.classic_vals
                                else final_mess[id]
                                for id in range(num_total)
                                if id not in new_reg.qubit_map}

        return (new_reg, final_mess)

    def apply_gate(self, gate, targets=None, controls=None, anticontrols=None,
                   num_threads=-1, target=None, control=None, anticontrol=None):
        """Add specified gate to this QGate.

        Positional arguments:
            gate: string with the name of the gate to apply, or a QGate
        Keyworded arguments:
            targets: id or list of ids of the qubits the gate will target
            controls: id or set of ids of the qubit that will act as controls
            anticontrols: id or set of ids of the qubit that will act as
                         anticontrols
            num_threads: number of threads to use
        Return:
            A new QRegistry
        """
        if target is not None:
            print("[WARNING] target keyworded argument is deprecated. Please use targets instead")
            if targets is not None:
                raise ValueError("target argument can't be set alongside targets")
            targets = target
        if control is not None:
            print("[WARNING] control keyworded argument is deprecated. Please use controls instead")
            if controls is not None:
                raise ValueError("control argument can't be set alongside controls")
            controls = control
        if anticontrol is not None:
            print("[WARNING] anticontrol keyworded argument is deprecated. Please use anticontrols instead")
            if anticontrols is not None:
                raise ValueError("anticontrol argument can't be set alongside anticontrols")
            anticontrols = anticontrol
        if not np.allclose(num_threads % 1, 0):
            raise ValueError("num_threads must be an integer")
        num_threads = int(num_threads)
        num_qubits = self.num_qubits + self.num_bits
        op_data = _get_op_data(num_qubits, 0, gate, targets, None, None,
                               controls, anticontrols, None, None)
        gate = op_data["gate"]
        targets = op_data["targets"]
        controls = op_data["controls"]
        anticontrols = op_data["anticontrols"]
        classic_controls = {qubit_id for qubit_id in controls
                            if qubit_id in self.classic_vals}
        classic_anticontrols = {qubit_id for qubit_id in anticontrols
                                if qubit_id in self.classic_vals}
        ccheck = all(self.classic_vals[id] for id in classic_controls)
        accheck = not any(self.classic_vals[id] for id in classic_anticontrols)
        if ((len(classic_controls) > 0 and not ccheck)
                or (len(classic_anticontrols) > 0 and not accheck)):
            return self.clone()
        controls -= classic_controls
        controls -= classic_anticontrols
        new_reg = None
        if type(gate) == SimpleGate:
            aux_targets = [self.qubit_map[id] for id in targets]
            aux_controls = {self.qubit_map[id] for id in controls}
            aux_anticontrols = {self.qubit_map[id] for id in anticontrols}
            auxg = gate.gate
            if hasattr(self.doki, "BRegistry"):
                auxg = gate.matrix
            doki_reg = self.doki.registry_apply(self.reg, auxg,
                                                aux_targets, aux_controls,
                                                aux_anticontrols,
                                                num_threads, self.verbose)
            new_reg = QRegistry(None, doki=self.doki)
            new_reg.reg = doki_reg
            new_reg.num_qubits = self.num_qubits
            new_reg.num_bits = self.num_bits
            new_reg.qubit_map = {k: v for k, v in self.qubit_map.items()}
            new_reg.classic_vals = {k: v for k, v in self.classic_vals.items()}
            new_reg.size = self.size
            new_reg.verbose = self.verbose
        else:
            try:
                new_reg, _ = apply_design(gate, self, {}, targets,
                                          controls, anticontrols,
                                          num_threads=num_threads)
            except Exception:
                raise ValueError("Gate must be a valid string or a QGate")
        return new_reg

    def __getitem__(self, key):
        """Return the selected items of the state."""
        return self.get_state(key=key)

    def clone(self, num_threads=-1):
        if not np.allclose(num_threads % 1, 0):
            raise ValueError("num_threads must be an integer")
        num_threads = int(num_threads)
        new_reg = QRegistry(None, doki=self.doki)
        new_reg.num_qubits = self.num_qubits
        new_reg.num_bits = self.num_bits
        new_reg.size = self.size
        if self.reg is not None:
            new_reg.reg = self.doki.registry_clone(self.reg, num_threads,
                                                   self.verbose)
        else:
            new_reg.reg = None
        new_reg.qubit_map = {k: v for k, v in self.qubit_map.items()}
        new_reg.classic_vals = {k: v for k, v in self.classic_vals.items()}
        return new_reg

    def get_state(self, key=None, canonical=False):
        """Return the selected items of the state."""
        start, stop, step = _get_key_with_defaults(key, self.size,
                                                   0, self.size, 1)
        res = np.array([self.doki.registry_get(self.reg, i, canonical,
                                               self.verbose)
                        for i in range(start, stop, step)])
        return res

    def density_matrix(self, canonical=False):
        """Return functional matrix of the density matrix."""
        return Funmatrix(self.doki.registry_density(self.reg, self.verbose),
                         "Rho")

    def get_entropy(self, **kwargs):
        """Calculate Von Newmann Entropy.

        Keyworded arguments:
            base: the base of the logarithm. Default = 2
                The string "e" is a valid value
        """
        base = kwargs.get('base', 2)
        entropy = 0
        size = self.get_state_size()
        rho = self.density_matrix()
        for i in range(size):
            p = rho[i, i].real
            if p > 0:
                if base == "e":
                    entropy += p * np.log(p)
                elif type(base) == int or type(base) == float:
                    entropy += p * np.log(p)/np.log(base)
        return -entropy

    def get_bloch_coords(self, key=None):
        """Get the polar coordinates of ONE qubit in the bloch sphere."""
        if hasattr(self.doki, "BRegistry"):
            start, stop, step = _get_key_with_defaults(key, self.num_qubits,
                                                       0, self.num_qubits, 1)
            ids = [id for id in range(start, stop, step)]
            coords = [(sp.acos(self.reg.r[id].v[2, 0]),
                       sp.atan2(self.reg.r[id].v[1, 0], self.reg.r[id].v[0, 0]))
                      if self.reg.r[id].v[0, 0] != 0
                      else (sp.acos(self.reg.r[id].v[2, 0]), 0)
                      for id in ids]
            return coords
        if self.num_qubits != 1:
            raise NotImplementedError("Bloch sphere is only supported for " +
                                      "1 qubit registries")
        if key is not None and key >= self.get_num_qubits():
            raise ValueError(f"Qubit {key} is not in this registry")
        if key is not None and key in self.classic_vals:
            raise ValueError("That is a classical bit, not a qubit")
        state = self.get_state(canonical=True)
        cos_t2 = state[0].real
        sin_t2 = abs(state[1])
        phi = arg(state[1])
        if phi == sp.nan:
            phi = 0
        theta = sp.atan2(sin_t2, cos_t2) * 2
        if theta < 0:
            theta += 2 * np.pi
        if phi < 0:
            phi += 2 * np.pi
        return (theta, phi)

    def bra(self):
        """Get the conjugated row form state vector (bra <v|)."""
        k = self.get_state()
        k.shape = (1, k.shape[0])
        return np.conjugate(k)

    def ket(self):
        """Get the column form state vector (ket |v>)."""
        k = self.get_state()
        k.shape = (k.shape[0], 1)
        return k

    def prob(self, id, num_threads=-1):
        """Get the odds of getting 1 when measuring specified qubit."""
        id = _get_qubit_set(self.get_num_qubits(), [id], True, "argument")[0]
        if id in self.classic_vals:
            return self.classic_vals[id]
        for i in range(self.get_num_qubits(), id):
            if i in self.classic_vals:
                id -= 1
        return self.doki.registry_prob(self.reg, id, num_threads, self.verbose)


def superposition(a, b, num_threads=-1, verbose=False):
    """Join two registries into one by calculating tensor product."""
    if not np.allclose(num_threads % 1, 0):
        raise ValueError("num_threads must be an integer")
    num_threads = int(num_threads)
    doki_reg = None
    if a.reg is None and b.reg is None:
        doki_reg = None
    elif a.reg is None:
        doki_reg = b.reg
    elif b.reg is None:
        doki_reg = a.reg
    else:
        doki_reg = a.doki.registry_join(a.reg, b.reg, num_threads, verbose)
    new_reg = QRegistry(None, doki=a.doki)
    new_reg.num_qubits = a.num_qubits + b.num_qubits
    new_reg.num_bits = a.num_bits + b.num_bits
    new_reg.size = 2**new_reg.num_qubits
    new_reg.reg = doki_reg
    b_total = b.num_qubits + b.num_bits
    new_keys = [k + b_total for k in a.qubit_map.keys()]
    new_keys += list(b.qubit_map.keys())
    new_keys.sort()
    new_reg.qubit_map = {new_keys[i]: i for i in range(len(new_keys))}
    new_reg.classic_vals = {k: v for k, v in b.classic_vals.items()}
    for k, v in a.classic_vals.items():
        new_reg.classic_vals[k + b_total] = v
    return new_reg



"""Module that executes a circuit using Forest (Rigetti API).

Not working yet
"""

'''
import pyquil as pq
import pyquil.gates as pqg


def execute(qcircuit, processor=None, iterations=1):
    qc = pq.get_qc
'''



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Handles design execution on target machine."""
import qsimov.connectors.qsimovapi as qapi
from qsimov.structures.qcircuit import QCircuit


class Drewom(object):
    """Handle QDesign execution on specified machine."""

    def __init__(self, qmachine="doki",
                 extra=None):
        """Initialize Drewom structure."""
        self.qmachine = qmachine
        if extra is None:
            extra = {}
        self.extra = extra

    def execute(self, qcircuit, iterations=1):
        """Execute given design in a machine iterations times."""
        if not isinstance(qcircuit, QCircuit):
            raise ValueError("qcircuit must be a QCircuit")
        if self.qmachine == "doki":
            return qapi.execute(qcircuit, iterations=iterations, **self.extra)
        else:
            raise NotImplementedError("Drewom does not support that machine")



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Module that simplify the way of working with Doki gates.

Data Structures:
    SimpleGate: Quantum gate with its matrix

Functions:
    add_gate: adds a gate to the list of available gates.
"""
import doki
import numpy as np
import qsimov.connectors.parser as prs
import sympy as sp

from collections.abc import Iterable
from qsimov.structures.qbase import QBase
from sympy.matrices import Matrix
from sympy.codegen.cfunctions import log2
from sympy.physics.quantum.dagger import Dagger


class SimpleGate(QBase):
    """Quantum gate with its associated matrix."""

    def __init__(self, gate_string):
        """Load a gate that is in the list of gates."""
        name, args, invert, self_invert = prs.get_gate_data(gate_string)
        self.matrix = _get_gate_matrix(name, args, self_invert)
        if self.matrix.shape[0] != self.matrix.shape[1]:
            raise ValueError("Not a square matrix")
        self.num_qubits = int(log2(self.matrix.shape[0]))
        if 2**self.num_qubits != self.matrix.shape[0]:
            raise ValueError("Shape must be 2^n x 2^n. n = number of qubits")
        inverse = self.matrix
        if not self_invert:
            inverse = self.matrix
            if isinstance(self.matrix, Matrix):
                inverse = Dagger(self.matrix)
            else:
                inverse = self.matrix.conj().T
        _check_inverse(gate_string, self.matrix, inverse)
        if invert:
            self.matrix = inverse
        aux = np.array(self.matrix).astype(complex)
        self.gate = doki.gate_new(self.num_qubits, aux.tolist(), False)
        self._str = name
        if args is not None:
            if prs._gate_data[name][2]:  # has_invert_arg
                args = args[:-1]
            self._str += "(" + ",".join([str(arg) for arg in args]) + ")"
        if invert and not self_invert:
            self._str += "-1"

    def __eq__(self, other):
        if not isinstance(other, SimpleGate):
            return False
        if self is other:
            return True
        return self._str == other._str

    def __str__(self):
        gate_str = self._str
        if gate_str[-2:] == "-1":
            gate_str = gate_str[:-2] + "\u2020"
        return gate_str

    def __repr__(self):
        return str(self)

    def get_doki_gate(self):
        return self.gate

    def get_matrix(self):
        return self.matrix

    def get_num_qubits(self):
        return self.num_qubits

    def invert(self):
        gate_str = self._str
        if gate_str[-2:] == "-1":
            gate_str = gate_str[:-2]
        else:
            gate_str += "-1"
        return SimpleGate(gate_str)


def _check_inverse(name, matrix, inverse):
    got = None
    if isinstance(matrix, Matrix):
        got = np.array(matrix @ inverse).astype(complex)
        '''
        if not ex.equals(got) and not ex.equals(sp.N(got)):
            print("Expected:", ex)
            print("Got:", got)
            print("Got:", sp.simplify(got))
            print(ex.equals(got))
            print([[ex[i, j].equals(got[i, j]) for j in range(ex.shape[1])] for i in range(ex.shape[0])])
            print(sp.N(got))
            print(ex.equals(sp.N(got)))
            raise ValueError("Failed testing inverse of " + name)
        '''
    else:
        got = np.dot(matrix, inverse)
    ex = np.eye(matrix.shape[0], dtype=complex)
    if not np.allclose(got, ex):
        raise ValueError("Failed testing inverse of " + name)


def _get_gate_matrix(name, args, self_invert):
    """Get gate matrix from parser's get_gate_data info."""
    gate_function = prs._gate_func[name]
    matrix = None
    if type(args) != str and isinstance(args, Iterable):
        matrix = gate_function(*args)
    elif args is not None:
        matrix = gate_function(args)
    else:
        matrix = gate_function()

    return Matrix(matrix)


def add_gate(name, funct, min_args, max_args, has_invert_arg=None,
             is_own_inverse=False, aliases=[], overwrite=False):
    """Add specified gate to the list of available gates."""
    if has_invert_arg is not None:
        print("[WARNING] has_invert_arg keyworded argument has been deprecated, and will be removed. Currently it does nothing.")
    if name in prs._gate_func and not overwrite:
        raise ValueError(name + " gate has already been defined. " +
                         "Set overwrite=True if you want to overwrite it.")
    aliases.append(name)
    aliases = {alias.lower() for alias in aliases}
    for alias in aliases:
        if alias in prs._gate_alias and prs._gate_alias[alias] != name:
            raise ValueError("Alias " + alias + " for gate " + name +
                             " is in use by gate " + prs._gate_alias[alias] +
                             ". Either remove this alias or modify the older" +
                             " gate.")
    prs._gate_func[name] = funct
    prs._gate_data[name] = (min_args, max_args, is_own_inverse)
    for alias in aliases:
        prs._gate_alias[alias.lower()] = name



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Module that executes a circuit using Doki (QSimov core).

This module provides functions that execute a given set of quantum gates
on a given quantum system, using Doki (the QSimov core written in C) to
perform the simulation.
"""
import numpy as np
from qsimov.structures.qstructure import QStructure, _get_op_data
from qsimov.structures.qdesign import QDesign


def _check_classical(classical_reg, c_controls, c_anticontrols):
    for k in c_controls:
        if classical_reg[k] is None:
            raise ValueError("Undefined value for classic " +
                             f"control bit {k}.")
    for k in c_anticontrols:
        if classical_reg[k] is None:
            raise ValueError("Undefined value for classic " +
                             f"anticontrol bit {k}.")
    return (all(classical_reg[k] for k in c_controls) and
            not any(classical_reg[k] for k in c_anticontrols))


def apply_design(qdesign, qstruct, classical_reg, targets=None,
                 c_targets=None, controls=None, anticontrols=None,
                 c_controls=None, c_anticontrols=None,
                 random_generator=np.random.rand, num_threads=-1):
    """Apply specified gate to specified qubit with specified controls.

    Positional arguments:
        qstruct: QRegistry or QSystem affected by this gate
    Keyworded arguments:
        targets: id or list of ids of the qubits the gate will target
        controls: id or set of ids of the qubit that will act as controls
        anticontrols: id or set of ids of the qubit that will act as
                     anticontrols
        num_threads: number of threads to use
    """
    if not isinstance(qdesign, QDesign):
        raise ValueError("qdesign must be a QGate or a QCircuit")
    if not isinstance(qstruct, QStructure):
        raise ValueError("qstruct must be a QRegistry or a QSystem")
    num_qubits = qstruct.get_num_qubits()
    num_bits = qdesign.get_num_bits()
    op_data = _get_op_data(num_qubits, num_bits, None,
                           targets, c_targets, None,
                           controls, anticontrols, c_controls, c_anticontrols,
                           empty=True)
    targets = op_data["targets"]
    c_targets = op_data["c_targets"]
    controls = op_data["controls"]
    anticontrols = op_data["anticontrols"]
    c_controls = op_data["c_controls"]
    c_anticontrols = op_data["c_anticontrols"]
    num_targets = qdesign.get_num_qubits()
    if len(targets) == 0:  # By default we use the least significant qubits
        targets = [i for i in range(num_targets)]
    if len(c_targets) == 0:  # By default we use the least significant qubits
        c_targets = [i for i in range(num_bits)]
    if len(targets) != num_targets:
        raise ValueError(f"Specified gate is for {num_targets} qubits." +
                         f" {len(targets)} qubit ids given")
    for k in c_controls:
        if (k not in classical_reg):
            raise ValueError("Undefined value for classic " +
                             f"control bit {k}.")
    for k in c_anticontrols:
        if (k not in classical_reg):
            raise ValueError("Undefined value for classic " +
                             f"anticontrol bit {k}.")
    if not _check_classical(classical_reg, c_controls, c_anticontrols):
        return (qstruct, classical_reg)
    new_struct = qstruct.clone()
    aux = None
    exception = None
    try:
        for gate_data in qdesign.get_operations():
            if gate_data == "BARRIER":
                continue
            curr_c_controls = {c_targets[i] for i in gate_data["c_controls"]}
            curr_c_acontrols = {c_targets[i]
                                for i in gate_data["c_anticontrols"]}
            if not _check_classical(classical_reg,
                                    curr_c_controls, curr_c_acontrols):
                continue
            aux = new_struct
            curr_targets = [targets[i] for i in gate_data["targets"]]
            curr_c_targets = [c_targets[i] for i in gate_data["c_targets"]]
            curr_controls = {targets[i] for i in gate_data["controls"]}
            curr_controls = curr_controls.union(controls)
            curr_anticontrols = {targets[i]
                                 for i in gate_data["anticontrols"]}
            curr_anticontrols = curr_anticontrols.union(anticontrols)
            # print(gate_data)
            # print(c_targets)
            curr_outputs = [c_targets[i] for i in gate_data["outputs"]]
            gate = gate_data["gate"]
            if isinstance(gate, QDesign):
                new_struct, _ = apply_design(gate, aux, classical_reg,
                                             targets=curr_targets,
                                             c_targets=curr_c_targets,
                                             controls=curr_controls,
                                             anticontrols=curr_anticontrols,
                                             random_generator=random_generator,
                                             num_threads=num_threads)
            elif gate == "MEASURE":
                new_struct, m = aux.measure(curr_targets,
                                            random_generator=random_generator)
                for i in range(len(curr_outputs)):
                    classical_reg[curr_outputs[i]] = m[curr_targets[i]]
            else:
                new_struct = aux.apply_gate(gate,
                                            targets=curr_targets,
                                            controls=curr_controls,
                                            anticontrols=curr_anticontrols,
                                            num_threads=num_threads)
            if aux is not qstruct:
                del aux
                aux = None
    except Exception as ex:
        exception = ex
        if aux is not None and aux is not qstruct:
            del aux
        if new_struct is not None:
            del new_struct
    if exception is not None:
        raise exception
    return (new_struct, classical_reg)


def execute(qcircuit, random_generator=np.random.rand, num_threads=-1,
            use_system=True, return_struct=False, core=None, iterations=1):
    """Execute the gates in lines on a qsystem.

    Gets repeated the specified number of iterations.
    Returns the result of each iteration.
    """
    from qsimov.structures.qsystem import QSystem
    from qsimov.structures.qregistry import QRegistry
    num_qubits = qcircuit.get_num_qubits()
    num_bits = qcircuit.get_num_bits()
    num_ancilla = len(qcircuit.ancilla)
    if core is None or core.upper() == "CPU":
        import doki
    elif core.upper() == "GPU":
        import doki_gpu as doki
    elif core.upper() == "MPI":
        import doki_mpi as doki
    elif isinstance(core, str):
        raise ValueError("Unknown Doki core. Valid values: CPU, GPU or MPI")
    else:
        doki = core
    this_struct = QSystem
    if not use_system:
        this_struct = QRegistry
    old_sys = this_struct(num_qubits, doki=doki)
    for i in range(len(qcircuit.ancilla)):
        if qcircuit.ancilla[i] == 1:
            aux = old_sys.apply_gate("X", targets=num_qubits-num_ancilla+i)
            del old_sys
            old_sys = aux
    results = []
    for i in range(iterations):
        classical_reg = [None for i in range(num_bits)]
        new_sys, _ = apply_design(qcircuit, old_sys, classical_reg,
                                  targets=[i for i in range(num_qubits)],
                                  c_targets=[i for i in range(num_bits)],
                                  random_generator=random_generator,
                                  num_threads=num_threads)
        if return_struct:
            results.append([new_sys, classical_reg])
        else:
            results.append(classical_reg)
            del new_sys
    del old_sys
    return results



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Module that provides a data structure representing a quantum circuit.

Data Structures:
    QCircuit: Quantum Circuit, built from gates and measurements
"""
from qsimov.structures.qregistry import QRegistry
from qsimov.structures.qstructure import _get_op_data
from qsimov.structures.qsystem import QSystem
from qsimov.structures.qdesign import QDesign
import qsimov.connectors.qsimovapi as qapi
import numpy as np


class QCircuit(QDesign):
    """Quantum Circuit, built from gates and measurements."""

    def __init__(self, num_qubits, num_bits, name, ancilla=None):
        """Quantum Circuit constructor.

        Positional arguments:
            num_qubits: maximum number of qubits affected by this circuit
            num_bits: maximum number of classical bits affected by this circuit
            name: name of the gate
        Keyworded arguments:
            ancilla: list of values of the ancilliary qubits (0 or 1)
        """
        if num_qubits is None or not np.allclose(num_qubits % 1, 0):
            raise ValueError("num_qubits must be an integer")
        num_qubits = int(num_qubits)
        if num_qubits <= 0:
            raise ValueError("num_qubits must be positive")
        if name is None:
            raise ValueError("name can't be None")
        if ancilla is None:
            ancilla = []
        aux = [int(val) for val in ancilla]
        if not np.allclose(aux, ancilla):
            raise ValueError("ancilla must be a list of integers")
        ancilla = aux
        if not all(0 <= val <= 1 for val in ancilla):
            raise ValueError("ancilla must be a list of 0 and 1")
        self.name = name
        self.ops = []
        self.ancilla = ancilla
        self.num_qubits = num_qubits
        self.num_bits = num_bits

    def draw(self):
        raise NotImplementedError("Draw has not been implemented yet")

    def get_num_qubits(self):
        return self.num_qubits

    def get_num_bits(self):
        return self.num_bits

    def get_operations(self):
        return self.ops

    def add_operation(self, gate, targets=None, c_targets=None, outputs=None,
                      controls=None, anticontrols=None,
                      c_controls=None, c_anticontrols=None,
                      target=None, c_target=None, output=None,
                      control=None, anticontrol=None,
                      c_control=None, c_anticontrol=None):
        """Apply specified operation to this QCircuit.

        Positional arguments:
            gate: string with the name of the gate to apply,
                  a QGate or a Measure
        Keyworded arguments:
            targets: id or list of ids of the qubits the gate will target
            controls: id or set of ids of the qubit that will act as controls
            anticontrols: id or set of ids of the qubit that will act as
                          anticontrols
            num_threads: number of threads to use
        """
        if target is not None:
            print("[WARNING] target keyworded argument is deprecated. Please use targets instead")
            if targets is not None:
                raise ValueError("target argument can't be set alongside targets")
            targets = target
        if output is not None:
            print("[WARNING] output keyworded argument is deprecated. Please use outputs instead")
            if outputs is not None:
                raise ValueError("output argument can't be set alongside outputs")
            outputs = output
        if control is not None:
            print("[WARNING] control keyworded argument is deprecated. Please use controls instead")
            if controls is not None:
                raise ValueError("control argument can't be set alongside controls")
            controls = control
        if anticontrol is not None:
            print("[WARNING] anticontrol keyworded argument is deprecated. Please use anticontrols instead")
            if anticontrols is not None:
                raise ValueError("anticontrol argument can't be set alongside anticontrols")
            anticontrols = anticontrol
        if c_control is not None:
            print("[WARNING] c_control keyworded argument is deprecated. Please use c_controls instead")
            if c_controls is not None:
                raise ValueError("c_control argument can't be set alongside c_controls")
            c_controls = c_control
        if c_anticontrol is not None:
            print("[WARNING] c_anticontrol keyworded argument is deprecated. Please use c_anticontrols instead")
            if c_anticontrols is not None:
                raise ValueError("c_anticontrol argument can't be set alongside c_anticontrols")
            c_anticontrols = c_anticontrol
        if gate is None:
            raise ValueError("Gate can't be None")
        if isinstance(gate, str) and gate.lower() == "barrier":
            self.ops.append("BARRIER")
            return
        aux = gate
        if isinstance(gate, str) and gate.lower() == "measure":
            aux = None
        op_data = _get_op_data(self.num_qubits, self.num_bits, aux, targets,
                               c_targets, outputs,
                               controls, anticontrols,
                               c_controls, c_anticontrols)
        if aux is None:
            op_data["gate"] = "MEASURE"
        self.ops.append(op_data)

    def execute(self, qubits, iterations=1, qmachine=None,
                args={"useSystem": True}, optimize=True):
        """Execute the circuit with specified configuration options.

        Positional arguments:
            qubits: list of 0 and 1 with the initial value of each
                        non ancilla qubit of the circuit
                OR  QRegistry or QSystem
        Keyworded arguments:
            iterations: the circuit will be executed this amount of times
            qmachine: string with the name of the machine in which the circuit
                      will be executed. Right now only supported QSimov
                      simulator (None or "local").
            args: dictionary with extra arguments for the specified qmachine.
                  For QSimov (local), "useSystem" is the only extra argument.
                  Defaults to True. Makes QSimov use QSystems instead of
                  QRegistries by default.
            optimize: whether or not to use oplines instead of lines.
                      Default = True
        """
        result = None
        lines = self.lines
        if optimize:
            lines = self.oplines
        if ((isinstance(qubits, QRegistry) or isinstance(qubits, QSystem))
                and iterations > 1):
            raise ValueError("Can not do more than one iteration with a " +
                             "precreated registry or system!")
        elif (qmachine is None or qmachine == "local"):
            result = qapi.execute(qubits, iterations, lines, self.ancilla,
                                  args["useSystem"], optimize)
        else:
            raise ValueError("Unsupported qmachine!")
        return result



#!/usr/bin/python
'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Deutsch-Jozsa with n qubit."""

from qsimov import Drewom, QCircuit, QGate
import sys
import time as t


def DJAlgCircuit(size, U_f):
    """Return Deutsch-Josza algorithm circuit.

    U_f is the oracle, having x1..xn and y as input qubits.
    x1..xn and y take only values either 0 or 1.
    Once applied it returns x1..xn and y', where y' = f(x1..xn) XOR y.
    Size is n + 1, where n is the number of input bits of f.
    """
    # The last qubit is defined as an ancilla qubit with value 1
    c = QCircuit(size, size-1, "Deutsch-Josza Algorithm", ancilla=[1])

    # We apply a Hadamard gate to all the qubits
    for i in range(size):
        c.add_operation("H", targets=i)

    # We apply the U_f oracle
    c.add_operation(U_f, targets=[i for i in range(size)])

    # We apply a Hadamard gate to all qubits but the last one
    for i in range(size - 1):
        c.add_operation("H", targets=i)

    # We measure x1..xn qubits.
    # If all of them are 0, f is constant, otherwise f is balanced.
    targets = [i for i in range(size - 1)]
    c.add_operation("MEASURE", targets=targets, outputs=targets)

    return c


def geq_zero(size):
    """DJ Oracle for f(x) = x >= 0. Balanced."""
    gate = QGate(size, 0, "U_(x>=0)")

    gate.add_operation("X", targets=size-1, anticontrols=size-2)

    return gate


def sample_main(nq, use_system=False, iterations=1):
    """Execute DJ with specified qubits."""
    # The number of qubits (x1..xn, y)
    gate = geq_zero(nq)  # The U_f oracle
    circuit = DJAlgCircuit(nq, gate)  # The Deutsch-Jozsa algorithm circuit
    # We specify useSystem = False to disable optimizations
    # other than Functional Matrices
    import numpy as np
    executor = Drewom(qmachine="doki",
                      extra={"num_threads": -1,
                             "random_generator": np.random.rand,
                             "use_system": use_system})
    init = t.time()
    mes = executor.execute(circuit, iterations=iterations)
    end = t.time()
    # print("RawMes:", mes)
    is_bal_iters = [any(iter[0][:-1]) for iter in mes]
    is_bal = all(is_bal_iters)
    if not is_bal and any(is_bal_iters):
        print("There is no consensus! BUG!")
    print("Is balanced?:", is_bal)
    print("Elapsed time:", end - init, "s")


if __name__ == "__main__":
    argv = sys.argv[1:]
    if 1 != len(argv) or int(argv[0]) < 2:  # We only have one argument
        print("Syntax: " + sys.argv[0] + " <number of qubits (min 2)>")
    else:
        sample_main(int(argv[0]), False, 1)



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from abc import abstractmethod
from qsimov.structures.qbase import QBase


class QDesign(QBase):
    @abstractmethod
    def add_operation(self, gate, targets=None, c_targets=None, outputs=None,
                      controls=None, anticontrols=None,
                      c_controls=None, c_anticontrols=None):
        pass

    @abstractmethod
    def get_operations(self):
        pass

    @abstractmethod
    def draw(self):
        pass

    @abstractmethod
    def get_num_bits(self):
        pass



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Module that provides a data structure representing a quantum gate.

Data Structures:
    QGate: Quantum Gate, built from elemental gates or other QGates
"""
import numpy as np
from qsimov.structures.qdesign import QDesign
from qsimov.structures.qstructure import _get_op_data


class QGate(QDesign):
    """Quantum Gate, built from elemental gates or other QGates."""

    def __init__(self, num_qubits, num_bits, name):
        """Quantum Gate constructor.

        Positional arguments:
            num_qubits: maximum number of qubits affected by this gate
            name: name of the gate
        """
        if num_qubits is None or not np.allclose(num_qubits % 1, 0):
            raise ValueError("num_qubits must be an integer")
        num_qubits = int(num_qubits)
        if num_qubits <= 0:
            raise ValueError("num_qubits must be positive")
        if name is None:
            raise ValueError("name can't be None")
        # List of lines. A line is a list of gates that can be executed
        # in parallel. One gate per qubit. None means Identity gate
        self.ops = []
        # Name of the gate
        self.name = str(name)
        # Maximum number of qubits affected by this gate
        self.num_qubits = num_qubits
        # Maximum number of bits affected by this gate
        self.num_bits = num_bits

    def __repr__(self):
        """Return string representation of the gate."""
        return self.name

    def __str__(self):
        """Return string representation of the gate."""
        return self.name

    def draw(self):
        raise NotImplementedError("Draw has not been implemented yet")

    def get_num_qubits(self):
        return self.num_qubits

    def get_num_bits(self):
        return self.num_bits

    def get_operations(self):
        return self.ops

    def add_operation(self, gate, targets=None, c_targets=None, outputs=None,
                      controls=None, anticontrols=None,
                      c_controls=None, c_anticontrols=None,
                      target=None, c_target=None, output=None,
                      control=None, anticontrol=None,
                      c_control=None, c_anticontrol=None):
        """Apply specified gate to specified qubit with specified controls.

        Positional arguments:
            comma separated gates, their sizes must match the number of
                qubits in the system. Sorted by their least significant
                target qubit id.
        Keyworded arguments:
            controls: id or set of ids of the qubits that will
                      act as controls
            anticontrols: id or set of ids of the qubits that will
                          act as anticontrols
            c_controls: id or set of ids of the classic bits that will
                        act as controls
            c_anticontrols: id set list of ids of the classic bits that will
                            act as anticontrols
        """
        if target is not None:
            print("[WARNING] target keyworded argument is deprecated. Please use targets instead")
            if targets is not None:
                raise ValueError("target argument can't be set alongside targets")
            targets = target
        if output is not None:
            print("[WARNING] output keyworded argument is deprecated. Please use outputs instead")
            if outputs is not None:
                raise ValueError("output argument can't be set alongside outputs")
            outputs = output
        if control is not None:
            print("[WARNING] control keyworded argument is deprecated. Please use controls instead")
            if controls is not None:
                raise ValueError("control argument can't be set alongside controls")
            controls = control
        if anticontrol is not None:
            print("[WARNING] anticontrol keyworded argument is deprecated. Please use anticontrols instead")
            if anticontrols is not None:
                raise ValueError("anticontrol argument can't be set alongside anticontrols")
            anticontrols = anticontrol
        if c_control is not None:
            print("[WARNING] c_control keyworded argument is deprecated. Please use c_controls instead")
            if c_controls is not None:
                raise ValueError("c_control argument can't be set alongside c_controls")
            c_controls = c_control
        if c_anticontrol is not None:
            print("[WARNING] c_anticontrol keyworded argument is deprecated. Please use c_anticontrols instead")
            if c_anticontrols is not None:
                raise ValueError("c_anticontrol argument can't be set alongside c_anticontrols")
            c_anticontrols = c_anticontrol
        if isinstance(gate, str) and gate.lower() == "barrier":
            self.ops.append("BARRIER")
            return
        if isinstance(gate, str) and gate.lower() == "measure":
            raise ValueError("A QGate can only contain gates or barriers")
        if outputs is not None:
            print("[WARNING] outputs argument is for measurements. Since they can't be used with QGate, do you mean targets?")
        num_qubits = self.num_qubits
        num_bits = self.num_bits
        op_data = _get_op_data(num_qubits, num_bits, gate, targets, c_targets,
                               outputs, controls, anticontrols,
                               c_controls, c_anticontrols)
        self.ops.append(op_data)

    def dagger(self):
        """Return the Conjugate Transpose of the given matrix."""
        return self.invert()

    def invert(self):
        """Return the Conjugate Transpose of the given matrix."""
        invgate = QGate(self.num_qubits, self.num_bits, self.name + "-1")
        for op_data in self.ops[::-1]:
            aux = op_data["gate"]
            if aux != "BARRIER":
                aux = aux.invert()
            invgate.add_operation(aux, targets=op_data["targets"],
                                  controls=op_data["controls"],
                                  anticontrols=op_data["anticontrols"],
                                  c_controls=op_data["c_controls"],
                                  c_anticontrols=op_data["c_anticontrols"])
        return invgate



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from qsimov.structures.funmatrix import Funmatrix
from qsimov.structures.qregistry import QRegistry, superposition
from qsimov.structures.qsystem import QSystem, join_systems
from qsimov.structures.qgate import QGate
from qsimov.structures.qcircuit import QCircuit
from qsimov.structures.simple_gate import SimpleGate, add_gate
from qsimov.connectors.drewom import Drewom
from qsimov.connectors.parser import get_available_gates, get_gate_aliases



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

import matplotlib.pyplot as plt
import numpy as np


def plot_arc3d(vector1, vector2, radius=0.2, ax=None, colour="C0", color=None):
    """ Plot arc between two given vectors in 3D space. """
    '''
    https://stackoverflow.com/questions/47321839/how-to-show-the-angle-by-an-arc-between-two-3d-vectors-in-matplotlib
    answer by Phil G (https://stackoverflow.com/users/3912166/phil-g)
    Calculate vector between two vector end points, and the resulting
    spherical angles for various points along this vector. From this,
    derive points that lie along the arc between vector1 and vector2.
    '''
    if color is not None:
        colour = color
    v = [i - j for i, j in zip(vector1, vector2)]
    v_points_direct = [(vector2[0] + v[0] * num,
                        vector2[1] + v[1] * num,
                        vector2[2] + v[2] * num) for num in np.linspace(0, 1)]

    if any([np.linalg.norm(v_point) == 0.0 for v_point in v_points_direct]):
        return
    v_points_aux = [(v_point, np.linalg.norm(v_point))
                    for v_point in v_points_direct]
    v_thetas = [np.arccos(v_point[2]/norm)
                for v_point, norm in v_points_aux if norm != 0.0]
    v_phis = [np.arctan2(v_point[1], v_point[0])
              for v_point, norm in v_points_aux if norm != 0.0]

    v_points_arc = [(radius*np.sin(theta)*np.cos(phi),
                     radius*np.sin(theta)*np.sin(phi),
                     radius*np.cos(theta))
                    for theta, phi in zip(v_thetas, v_phis)]
    v_points_arc.append((0, 0, 0))

    from mpl_toolkits.mplot3d.art3d import Poly3DCollection
    """ Plot polygon.
    Face colour must be set afterwards, otherwise it over-rides
    the transparency.
    https://stackoverflow.com/questions/18897786/transparency-for-poly3dcollection-plot-in-matplotlib
    """
    points_collection = Poly3DCollection([v_points_arc], alpha=0.4)
    points_collection.set_facecolor(colour)
    ax.add_collection3d(points_collection)


def draw_bloch_sphere(theta, phi, length=1.0):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    plt.axis("off")
    # ax.set_aspect("equal")  # NotImplementedError
    # Alternative to equal aspect:
    ax.set_box_aspect([ub - lb
                       for lb, ub in (getattr(ax, f"get_{a}lim")()
                                      for a in "xyz")])

    # draw axes
    # X axis
    ax.quiver(-1.1, 0, 0, 2.2, 0, 0, color="r",
              arrow_length_ratio=0.07, linewidths=0.75)
    # Y axis
    ax.quiver(0, -1.1, 0, 0, 2.2, 0, color="g",
              arrow_length_ratio=0.07, linewidths=0.75)
    # Z axis
    ax.quiver(0, 0, -1.1, 0, 0, 2.2, color="b",
              arrow_length_ratio=0.07, linewidths=0.75)

    # ax.plot([-1.1, 1.1],[0, 0], [0, 0],lw=1,color="r")  # X axis
    # ax.plot([0, 0],[-1.1, 1.1], [0, 0],lw=1,color="g")  # Y axis
    # ax.plot([0, 0], [0, 0],[-1.1, 1.1],lw=1,color="b")  # Z axis
    ax.text(-1.25, 0, 0, "|-\u27E9", fontsize=12,
            horizontalalignment="center", verticalalignment="center")
    ax.text(+1.25, 0, 0, "|+\u27E9", fontsize=12,
            horizontalalignment="center", verticalalignment="center")
    ax.text(0, -1.25, 0, "|-i\u27E9", fontsize=12,
            horizontalalignment="center", verticalalignment="center")
    ax.text(0, +1.25, 0, "|i\u27E9", fontsize=12,
            horizontalalignment="center", verticalalignment="center")
    ax.text(0, 0, -1.25, "|1\u27E9", fontsize=12,
            horizontalalignment="center", verticalalignment="center")
    ax.text(0, 0, +1.25, "|0\u27E9", fontsize=12,
            horizontalalignment="center", verticalalignment="center")
    ax.scatter([0], [0], [0], color="k", s=10)  # Origin

    # draw sphere
    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)

    x = 1 * np.outer(np.cos(u), np.sin(v))
    y = 1 * np.outer(np.sin(u), np.sin(v))
    z = 1 * np.outer(np.ones(np.size(u)), np.cos(v))
    ax.plot_surface(x, y, z, rstride=4, cstride=4, color="white",
                    linewidth=0, alpha=0.25)

    # Calculate the qubit point
    x = np.cos(phi) * np.sin(theta) * length
    y = np.sin(phi) * np.sin(theta) * length
    z = np.cos(theta) * length
    # draw a point
    ax.scatter([x], [y], [z], color="k", s=10)
    '''
    coord_string = f"({round(x, 2)}, {round(y, 2)}, {round(z, 2)}"
    if length != 1.0:
        coord_string += f", {round(length, 2)})"
    coord_string += ")"
    ax.text(x * 1.25, y * 1.25, z * 1.25, coord_string, fontsize=12,
            horizontalalignment="center", verticalalignment="center")
    '''
    # draw a vector
    ax.quiver(0, 0, 0, x, y, z, color="k", linewidths=2)
    # draw dashed lines
    ax.plot([x, x], [y, y], [z, 0], lw=1, color="k", linestyle="dashed")
    ax.plot([x, 0], [y, 0], [0, 0], lw=1, color="k", linestyle="dashed")
    # draw angles
    # theta
    plot_arc3d([0, 0, 1], [x, y, z], radius=0.25*length, ax=ax, color="b")

    # phi
    if phi < np.pi:
        plot_arc3d([1, 0, 0], [x, y, 0], radius=0.25*length, ax=ax, color="r")
    elif phi >= np.pi:
        plot_arc3d([1, 0, 0], [0, 1, 0], radius=0.25*length, ax=ax, color="r")
        plot_arc3d([0, 1, 0], [-1, 0, 0], radius=0.25*length, ax=ax, color="r")
        plot_arc3d([-1, 0, 0], [x, y, 0], radius=0.25*length, ax=ax, color="r")

    ax.view_init(elev=15, azim=15)
    # plt.show()
    return fig


# draw_bloch_sphere(3*np.pi/4, 3.5*np.pi/2, length=1)



'''
QSimov: A Quantum Computing Toolkit.
Copyright (C) 2017  Hernán Indíbil de la Cruz Calvo

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

"""Module that provides a data structure representing a quantum system.

Data Structures:
    QSystem: Quantum System, preferred over QRegistry (can save a lot of space)

Functions:
    superposition: join two registries into one by calculating tensor product.
"""
import numpy as np
from qsimov.structures.qstructure import QStructure, _get_qubit_set, \
                                        _get_op_data, _get_key_with_defaults
from qsimov.structures.qregistry import QRegistry, superposition


class QSystem(QStructure):
    """Quantum System, preferred over QRegistry (can save a lot of space)."""

    def __init__(self, num_qubits, data=None, doki=None, verbose=False):
        """Initialize QSystem to state 0.

        num_qubits -> number of QuBits in the system.
        """
        if doki is None:
            import doki
        self.doki = doki
        if data is None:
            if num_qubits is None:
                self.regs = None
                self.qubitMap = None
                self.num_qubits = 0
            else:
                self.regs = [[QRegistry(1, doki=self.doki), [id]]
                             for id in range(num_qubits)]
                self.qubitMap = {id: id for id in range(num_qubits)}
                self.num_qubits = num_qubits
        else:
            self.regs = [[QRegistry(None, data=data["regs"][i][0], doki=self.doki),
                          data["regs"][i][1]]
                         for i in range(len(data["regs"]))]
            self.qubitMap = data["qubit_map"]
            self.num_qubits = data["num_qubits"]
        self.verbose = verbose

    def get_data(self):
        return {"regs": [[self.regs[i][0].get_data(), self.regs[i][1]]
                         for i in range(len(self.regs))],
                "qubit_map": self.qubitMap,
                "num_qubits": self.num_qubits}

    def free(self, deep=False):
        """Release memory held by the QSystem."""
        if self.regs is not None:
            if deep:
                for reg, _ in self.regs:
                    if isinstance(reg, QRegistry):
                        reg.free()
            del self.regs
            del self.qubitMap
            self.regs = None
            self.qubitMap = None

    def clone(self, deep=False):
        """Clone this QSystem."""
        new_sys = QSystem(None, doki=self.doki)
        new_sys.num_qubits = self.num_qubits
        new_sys.qubitMap = {}
        for id in self.qubitMap:
            new_sys.qubitMap[id] = self.qubitMap[id]
        new_sys.regs = [[self.regs[id][0],
                         self.regs[id][1][:]]
                        if not deep else
                        [self.regs[id][0].clone(),
                         self.regs[id][1][:]]
                        for id in range(len(self.regs))]
        return new_sys

    def __del__(self):
        """Clean after deletion."""
        self.free()

    def prob(self, id, num_threads=-1):
        """Get the odds of getting 1 when measuring specified qubit."""
        id = _get_qubit_set(self.get_num_qubits(), [id], True, "argument")[0]
        reg, ids = self.regs[self.qubitMap[id]]
        new_id = None
        for i in range(len(ids)):
            if ids[i] == id:
                new_id = i
                break
        if new_id is None:
            raise RuntimeError("Couldn't find id in any reg, " +
                               "please report this bug.")
        return reg.prob(new_id, num_threads=num_threads)

    def get_sizes(self):
        """Return the number of elements of each registry in the system."""
        return ((reg[0].get_state_size(), reg[1])
                if type(reg[0]) == QRegistry
                else (1, reg[1])
                for reg in self.regs)

    def get_state_size(self):
        """Return the number of elements in the state vector of the system."""
        total = 0
        for reg in self.regs:
            if type(reg[0]) == QRegistry:
                total += reg[0].get_state_size()
            else:
                total += 1
        return total

    def get_split_num_qubits(self):
        """Return the number of qubits in each registry of the system."""
        return (reg[0].get_num_qubits()
                if type(reg[0]) == QRegistry
                else 1  # When we measure with remove=True
                for reg in self.regs)

    def get_num_qubits(self):
        """Return the number of qubits in this system."""
        return self.num_qubits

    def measure(self, ids, random_generator=np.random.rand,
                num_threads=-1, deep=False):
        """Measure specified qubits of this system and collapse.

        Positional arguments:
            ids -> List of QuBit ids that have to be measured
        Keyworded arguments:
            random_generator -> function without arguments that returns
                                a random real number in [0, 1)
        Return:
            List with the value obtained after each measure
        """
        num_qubits = self.get_num_qubits()
        ids = _get_qubit_set(num_qubits, ids, False, "ids")
        if ids is None:
            raise ValueError("ids cannot be None")
        split_ids = {reg_id: set() for reg_id in range(len(self.regs))}
        for qubit_id in ids:
            reg_id = self.qubitMap[qubit_id]
            split_ids[reg_id].add(qubit_id)
        # In split ids we have reg_id -> set of ids to measure in that reg
        split_ids = {k: v for k, v in split_ids.items() if len(v) > 0}
        result = [None for i in range(num_qubits)]
        # Here we have the registries that have not been used
        untouched_regs = {i for i in range(len(self.regs))
                          if i not in split_ids}
        # We create a new QSystem with the regs that have not been used
        new_sys = QSystem(None, doki=self.doki)
        new_sys.regs = []
        new_sys.qubitMap = {}
        exception = None
        try:
            for reg_id in untouched_regs:
                reggie, reg_ids = self.regs[reg_id]
                if deep:
                    reggie = reggie.clone()
                    reg_ids = reg_ids[:]
                new_sys.regs.append((reggie, reg_ids))
            for reg_id in range(len(untouched_regs)):
                for qubit_id in new_sys.regs[reg_id][1]:
                    new_sys.qubitMap[qubit_id] = reg_id
            new_sys.num_qubits = self.num_qubits

            # We iterate through the registries that have a qubit in ids
            for reg_id in split_ids:
                partial_ids = split_ids[reg_id]  # ids of QSystem to measure
                new_reg = None
                partial_result = None
                reg, reg_ids = self.regs[reg_id]
                # Ids to measure in the QRegistry (not in the whole QSystem)
                # mapped to the id in the QSystem
                new_ids = {i: reg_ids[i] for i in range(len(reg_ids))
                           if reg_ids[i] in partial_ids}
                # Not measured ids in this registry
                new_reg_ids = [id for id in reg_ids if id not in partial_ids]
                # We measure registries
                aux = reg.measure(new_ids.keys(),
                                  random_generator=random_generator,
                                  num_threads=num_threads)
                new_reg, partial_result = aux
                # We add the results to the result list
                for local_id in new_ids:
                    result[new_ids[local_id]] = partial_result[local_id]
                # We add the new registry to the list of regs
                if len(new_reg_ids) > 0:
                    aux_id = len(new_sys.regs)
                    new_sys.regs.append((new_reg, reg_ids))
                    for id in new_reg_ids:
                        new_sys.qubitMap[id] = aux_id
                else:
                    new_reg.free()
                # We add new registries with only the measured qubits
                for id in partial_ids:
                    one_reg = QRegistry(1, doki=self.doki,
                                        verbose=self.verbose)
                    if result[id]:
                        one_aux = one_reg.apply_gate("X")
                        one_reg.free()
                        one_reg = one_aux
                    new_sys.regs.append((one_reg, [id]))
                    new_sys.qubitMap[id] = len(new_sys.regs) - 1
        except Exception as ex:
            exception = ex
        if exception is not None:
            del new_sys
            raise exception
        return (new_sys, result)

    def as_qregistry(self, num_threads=-1, canonical=False):
        """Return this system as a QRegistry."""
        aux_reg = None
        new_reg = None
        new_ids = []
        first = True
        for reg_id in range(len(self.regs)):
            reg, ids = self.regs[reg_id]
            if new_reg is None:
                new_reg = reg
            else:
                aux_reg = superposition(new_reg, reg,
                                        num_threads=num_threads,
                                        verbose=self.verbose)
            new_ids = ids + new_ids
            if aux_reg is not None:
                if not first:
                    del new_reg
                first = False
                new_reg = aux_reg
                aux_reg = None
        # Here we remove the unused ids
        q_ids = [id for id in new_ids if new_reg.get_classic(id) is None]
        swap_ids = np.argsort(np.argsort(q_ids))
        # And we sort the remaining qubits by qubit_id
        for i in range(len(swap_ids)):
            while swap_ids[i] != i:
                swap_targets = [swap_ids[i], swap_ids[swap_ids[i]]]
                swap_ids[swap_targets[0]], swap_ids[i] = swap_targets
                aux_reg = new_reg.apply_gate("SWAP",
                                             targets=[i, swap_targets[0]],
                                             num_threads=num_threads)
                if not first:
                    del new_reg
                new_reg = aux_reg
        return new_reg

    def get_state(self, key=None, canonical=False):
        return self.as_qregistry().get_state(key=key, canonical=canonical)

    def get_classic(self, id):
        """Return classic bit value."""
        return None

    def apply_gate(self, gate, targets=None, controls=None, anticontrols=None,
                   num_threads=-1, deep=False, target=None, control=None, anticontrol=None):
        """Apply specified gate to specified qubit with specified controls.

        Positional arguments:
            gate: string with the name of the gate to apply, or a QGate
        Keyworded arguments:
            targets: id of the least significant qubit the gate will target
            controls: id or list of ids of the qubit that will act as
                     controls
            anticontrols: id or list of ids of the qubit that will act as
                         anticontrols
            num_threads: number of threads to use
            optimize: only for QGates. Whether to use optimized lines or
                      user defined lines
        """
        if target is not None:
            print("[WARNING] target keyworded argument is deprecated. Please use targets instead")
            if targets is not None:
                raise ValueError("target argument can't be set alongside targets")
            targets = target
        if control is not None:
            print("[WARNING] control keyworded argument is deprecated. Please use controls instead")
            if controls is not None:
                raise ValueError("control argument can't be set alongside controls")
            controls = control
        if anticontrol is not None:
            print("[WARNING] anticontrol keyworded argument is deprecated. Please use anticontrols instead")
            if anticontrols is not None:
                raise ValueError("anticontrol argument can't be set alongside anticontrols")
            anticontrols = anticontrol
        if not np.allclose(num_threads % 1, 0):
            raise ValueError("num_threads must be an integer")
        num_threads = int(num_threads)
        num_qubits = self.get_num_qubits()
        op_data = _get_op_data(num_qubits, 0, gate, targets, None, None,
                               controls, anticontrols, None, None)
        gate = op_data["gate"]
        targets = op_data["targets"]
        controls = op_data["controls"]
        anticontrols = op_data["anticontrols"]
        # We create a new system without the data of the parties
        new_sys = QSystem(None, doki=self.doki)
        new_sys.regs = []
        new_reg = None
        aux_reg = None
        exception = None
        try:
            # If any of the affected qubits is marked as not usable
            if any([self.get_classic(qubit_id) is not None
                    for qubit_id in targets]):
                # we raise an exception
                raise ValueError("Trying to apply gate to classic bit")
            cfail = any([self.get_classic(qubit_id) is False
                         for qubit_id in controls])
            acfail = any([self.get_classic(qubit_id) is True
                          for qubit_id in anticontrols])
            if cfail or acfail:
                if deep:
                    return self.clone(deep=True)
                else:
                    return self
            controls = {qubit_id for qubit_id in controls
                        if self.get_classic(qubit_id) is None}
            anticontrols = {qubit_id for qubit_id in anticontrols
                            if self.get_classic(qubit_id) is None}
            # All affected qubits
            parties = controls.union(anticontrols).union(targets)
            touched_regs = {self.qubitMap[qubit_id]
                            for qubit_id in parties}
            for reg_id in range(len(self.regs)):
                if reg_id not in touched_regs:
                    reggie, reg_ideses = self.regs[reg_id]
                    if deep and isinstance(reggie, QRegistry):
                        reggie = reggie.clone()
                    new_sys.regs.append([reggie, reg_ideses[:]])
            # Create new qubit map
            new_sys.qubitMap = {}
            for reg_id in range(len(new_sys.regs)):
                for qubit_id in new_sys.regs[reg_id][1]:
                    new_sys.qubitMap[qubit_id] = reg_id
            new_sys.num_qubits = self.num_qubits
            new_ids = []
            merged = False
            for reg_id in touched_regs:
                curr_reg, curr_ids = self.regs[reg_id]
                if new_reg is not None:
                    aux_reg = superposition(curr_reg, new_reg,
                                            num_threads=num_threads,
                                            verbose=self.verbose)
                    if merged:
                        del new_reg
                    else:
                        merged = True
                    new_reg = aux_reg
                else:
                    new_reg = curr_reg
                new_ids += curr_ids
            inverse_map = {new_ids[qubit_id]: qubit_id
                           for qubit_id in range(len(new_ids))}
            mapped_targets = [inverse_map[qubit_id]
                              for qubit_id in targets]
            mapped_controls = {inverse_map[qubit_id]
                               for qubit_id in controls}
            mapped_anticontrols = {inverse_map[qubit_id]
                                   for qubit_id in anticontrols}
            aux_reg = new_reg.apply_gate(gate, targets=mapped_targets,
                                         controls=mapped_controls,
                                         anticontrols=mapped_anticontrols,
                                         num_threads=num_threads)
            if merged:
                del new_reg
                new_reg = None
            new_sys.regs.append([aux_reg, new_ids])
            for id in new_ids:
                new_sys.qubitMap[id] = len(new_sys.regs) - 1
        except Exception as ex:
            if new_sys is not None:
                del new_sys
            if new_reg is not None and merged:
                del new_reg
            if aux_reg is not None:
                del aux_reg
            new_sys = None
            exception = ex
        if exception is not None:
            raise exception
        return new_sys

    def get_bloch_coords(self, key=None):
        """Get the polar coordinates of all ONE qubit registries."""
        start, stop, step = _get_key_with_defaults(key, self.num_qubits,
                                                   0, self.num_qubits, 1)
        ids = [id for id in range(start, stop, step)]
        coords = [None for id in ids]
        for i in range(len(ids)):
            id = ids[i]
            try:
                reg, ids = self.regs[self.qubitMap[id]]
                new_id = ids.index(id)
                coords[i] = reg.get_bloch_coords(new_id)
            except Exception:
                pass
        if key is not None and type(key) != slice:
            coords = coords[0]
        return coords


def join_systems(most, least, deep=False):
    """Return a system that contains both a and b systems."""
    res = QSystem(None, doki=most.doki)
    res.regs = []
    res.qubitMap = {}
    exception = None
    try:
        count = 0
        for reg, ids in least.regs:
            new_reg = reg
            if reg == QRegistry:
                if deep:
                    new_reg = reg.clone()
            count += 1
            res.regs.append([new_reg, ids[:]])
        offset = least.get_num_qubits()
        for reg, ids in most.regs:
            new_reg = reg
            if reg == QRegistry:
                if deep:
                    new_reg = reg.clone()
            count += 1
            res.regs.append([new_reg, [id + offset for id in ids]])
        for i in range(len(res.regs)):
            _, ids = res.regs[i]
            for qubit_id in ids:
                res.qubitMap[qubit_id] = i
    except Exception as ex:
        exception = ex
    if exception is not None:
        del res
        raise exception
    return res
